# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""entity classes for optional library entities"""


from warnings import warn
from datetime import datetime

from six.moves import range
from six.moves.urllib.parse import urlsplit, urlunsplit

from logilab.mtconverter import xml_escape

from cubicweb import UnknownProperty
from cubicweb.entity import _marker
from cubicweb.entities import AnyEntity, fetch_config

def mangle_email(address):
    try:
        name, host = address.split('@', 1)
    except ValueError:
        return address
    return '%s at %s' % (name, host.replace('.', ' dot '))


class EmailAddress(AnyEntity):
    __regid__ = 'EmailAddress'
    fetch_attrs, cw_fetch_order = fetch_config(['address', 'alias'])
    rest_attr = 'eid'

    def dc_title(self):
        if self.alias:
            return '%s <%s>' % (self.alias, self.display_address())
        return self.display_address()

    @property
    def email_of(self):
        return self.reverse_use_email and self.reverse_use_email[0] or None

    @property
    def prefered(self):
        return self.prefered_form and self.prefered_form[0] or self

    def related_emails(self, skipeids=None):
        # XXX move to eemail
        # check email relations are in the schema first
        subjrels = self.e_schema.object_relations()
        if not ('sender' in subjrels and 'recipients' in subjrels):
            return
        rset = self._cw.execute('DISTINCT Any X, S, D ORDERBY D DESC '
                                'WHERE X sender Y or X recipients Y, '
                                'X subject S, X date D, Y eid %(y)s',
                                {'y': self.eid})
        if skipeids is None:
            skipeids = set()
        for i in range(len(rset)):
            eid = rset[i][0]
            if eid in skipeids:
                continue
            skipeids.add(eid)
            yield rset.get_entity(i, 0)

    def display_address(self):
        if self._cw.vreg.config['mangle-emails']:
            return mangle_email(self.address)
        return self.address

    def printable_value(self, attr, value=_marker, attrtype=None,
                        format='text/html'):
        """overriden to return displayable address when necessary"""
        if attr == 'address':
            address = self.display_address()
            if format == 'text/html':
                address = xml_escape(address)
            return address
        return super(EmailAddress, self).printable_value(attr, value, attrtype, format)


class Bookmark(AnyEntity):
    """customized class for Bookmark entities"""
    __regid__ = 'Bookmark'
    fetch_attrs, cw_fetch_order = fetch_config(['title', 'path'])

    def actual_url(self):
        url = self._cw.build_url(self.path)
        if self.title:
            urlparts = list(urlsplit(url))
            if urlparts[3]:
                urlparts[3] += '&vtitle=%s' % self._cw.url_quote(self.title)
            else:
                urlparts[3] = 'vtitle=%s' % self._cw.url_quote(self.title)
            url = urlunsplit(urlparts)
        return url

    def action_url(self):
        return self.absolute_url() + '/follow'


class CWProperty(AnyEntity):
    __regid__ = 'CWProperty'

    fetch_attrs, cw_fetch_order = fetch_config(['pkey', 'value'])
    rest_attr = 'pkey'

    def typed_value(self):
        return self._cw.vreg.typed_value(self.pkey, self.value)

    def dc_description(self, format='text/plain'):
        try:
            return self._cw._(self._cw.vreg.property_info(self.pkey)['help'])
        except UnknownProperty:
            return u''


class CWCache(AnyEntity):
    """Cache"""
    __regid__ = 'CWCache'
    fetch_attrs, cw_fetch_order = fetch_config(['name'])

    def __init__(self, *args, **kwargs):
        warn('[3.19] CWCache entity type is going away soon. '
             'Other caching mechanisms can be used more reliably '
             'to the same effect.',
             DeprecationWarning)
        super(CWCache, self).__init__(*args, **kwargs)

    def touch(self):
        self._cw.execute('SET X timestamp %(t)s WHERE X eid %(x)s',
                         {'t': datetime.now(), 'x': self.eid})

    def valid(self, date):
        if date:
            return date > self.timestamp
        return False



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Specific views for schema related entities"""


from cubicweb import _

from itertools import cycle

import tempfile
import os, os.path as osp
import codecs

from six import text_type

from logilab.common.graph import GraphGenerator, DotBackend
from logilab.common.ureports import Section, Table
from logilab.common.registry import yes
from logilab.mtconverter import xml_escape
from yams import BASE_TYPES, schema2dot as s2d
from yams.buildobjs import DEFAULT_ATTRPERMS

from cubicweb.predicates import (is_instance, match_user_groups, match_kwargs,
                                has_related_entities, authenticated_user)
from cubicweb.schema import (META_RTYPES, SCHEMA_TYPES, SYSTEM_RTYPES,
                             WORKFLOW_TYPES, INTERNAL_TYPES)
from cubicweb.utils import make_uid
from cubicweb.view import EntityView, StartupView
from cubicweb import tags, uilib
from cubicweb.web import action, facet, schemaviewer
from cubicweb.web.views import uicfg, primary, baseviews, tabs, tableview, ibreadcrumbs

ALWAYS_SKIP_TYPES = BASE_TYPES | SCHEMA_TYPES
SKIP_TYPES  = (ALWAYS_SKIP_TYPES | META_RTYPES | SYSTEM_RTYPES | WORKFLOW_TYPES
               | INTERNAL_TYPES)
SKIP_TYPES.update(set(('CWUser', 'CWGroup', 'EmailAddress', 'Bookmark')))

def skip_types(req):
    if int(req.form.get('skipmeta', True)):
        return SKIP_TYPES
    return ALWAYS_SKIP_TYPES

_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl

for _action in ('read', 'add', 'update', 'delete'):
    _pvs.tag_subject_of(('*', '%s_permission' % _action, '*'), 'hidden')
    _pvs.tag_object_of(('*', '%s_permission' % _action, '*'), 'hidden')

for _etype in ('CWEType', 'CWRType', 'CWAttribute', 'CWRelation'):
    _pvdc.tag_attribute((_etype, 'description'), {'showlabel': False})

_pvs.tag_attribute(('CWEType', 'name'), 'hidden')
_pvs.tag_attribute(('CWEType', 'final'), 'hidden')
_pvs.tag_object_of(('*', 'workflow_of', 'CWEType'), 'hidden')
_pvs.tag_subject_of(('CWEType', 'default_workflow', '*'), 'hidden')
_pvs.tag_object_of(('*', 'specializes', 'CWEType'), 'hidden')
_pvs.tag_subject_of(('CWEType', 'specializes', '*'), 'hidden')
_pvs.tag_object_of(('*', 'from_entity', 'CWEType'), 'hidden')
_pvs.tag_object_of(('*', 'to_entity', 'CWEType'), 'hidden')

_pvs.tag_attribute(('CWRType', 'name'), 'hidden')
_pvs.tag_attribute(('CWRType', 'final'), 'hidden')
_pvs.tag_object_of(('*', 'relation_type', 'CWRType'), 'hidden')

_pvs.tag_subject_of(('CWAttribute', 'constrained_by', '*'), 'hidden')
_pvs.tag_subject_of(('CWRelation', 'constrained_by', '*'), 'hidden')


class SecurityViewMixIn(object):
    """mixin providing methods to display security information for a entity,
    relation or relation definition schema
    """
    cssclass = "listing schemaInfo"

    def permissions_table(self, erschema, permissions=None):
        self._cw.add_css('cubicweb.acl.css')
        w = self.w
        _ = self._cw._
        w(u'<table class="%s">' % self.cssclass)
        w(u'<tr><th>%s</th><th>%s</th><th>%s</th></tr>' % (
            _("permission"), _('granted to groups'), _('rql expressions')))
        for action in erschema.ACTIONS:
            w(u'<tr><td>%s</td><td>' % _(action))
            if permissions is None:
                groups = erschema.get_groups(action)
                rqlexprs = sorted(e.expression for e in erschema.get_rqlexprs(action))
            else:
                groups = permissions[action][0]
                rqlexprs = permissions[action][1]
            # XXX get group entity and call it's incontext view
            groups = [u'<a class="%s" href="%s">%s</a>' % (
                group, self._cw.build_url('cwgroup/%s' % group), label)
                      for label, group in sorted((_(g), g) for g in groups)]
            w(u'<br/>'.join(groups))
            w(u'</td><td>')
            w(u'<br/>'.join(rqlexprs))
            w(u'</td></tr>\n')
        w(u'</table>')

    def grouped_permissions_table(self, rschema):
        # group relation definitions with identical permissions
        perms = {}
        for rdef in rschema.rdefs.values():
            rdef_perms = []
            for action in rdef.ACTIONS:
                groups = sorted(rdef.get_groups(action))
                exprs = sorted(e.expression for e in rdef.get_rqlexprs(action))
                rdef_perms.append( (action, (tuple(groups), tuple(exprs))) )
            rdef_perms = tuple(rdef_perms)
            if rdef_perms in perms:
                perms[rdef_perms].append( (rdef.subject, rdef.object) )
            else:
                perms[rdef_perms] = [(rdef.subject, rdef.object)]
        # set layout permissions in a table for each group of relation
        # definition
        w = self.w
        _ = self._cw._
        w(u'<div style="margin: 0px 1.5em">')
        tmpl = u'<strong>%s</strong> %s <strong>%s</strong>'
        for perm, rdefs in perms.items():
            w(u'<div>%s</div>' % u', '.join(
                tmpl % (_(s.type), _(rschema.type), _(o.type)) for s, o in rdefs))
            # accessing rdef from previous loop by design: only used to get
            # ACTIONS
            self.permissions_table(rdef, dict(perm))
        w(u'</div>')


# global schema view ###########################################################

class SchemaView(tabs.TabsMixin, StartupView):
    """display schema information (graphically, listing tables...) in tabs"""
    __regid__ = 'schema'
    title = _('data model schema')
    tabs = [_('schema-diagram'), _('schema-entity-types'),
            _('schema-relation-types')]
    default_tab = 'schema-diagram'

    def call(self):
        self.w(u'<h1>%s</h1>' % self._cw._(self.title))
        self.render_tabs(self.tabs, self.default_tab)


class SchemaImageTab(StartupView):
    __regid__ = 'schema-diagram'

    def call(self):
        _ = self._cw._
        self.w(self._cw._(
            u'<div>This schema of the data model <em>excludes</em> the '
            'meta-data, but you can also display a <a href="%s">complete '
            'schema with meta-data</a>.</div>')
               % xml_escape(self._cw.build_url('view', vid='schemagraph', skipmeta=0)))
        self.w(u'<div><a href="%s">%s</a></div>' %
               (self._cw.build_url('view', vid='owl'),
                self._cw._(u'Download schema as OWL')))
        self.wview('schemagraph')

class SchemaETypeTab(StartupView):
    __regid__ = 'schema-entity-types'

    def call(self):
        self.wview('table', self._cw.execute(
            'Any X ORDERBY N WHERE X is CWEType, X name N, X final FALSE'))


class SchemaRTypeTab(StartupView):
    __regid__ = 'schema-relation-types'

    def call(self):
        self.wview('table', self._cw.execute(
            'Any X ORDERBY N WHERE X is CWRType, X name N, X final FALSE'))

# CWEType ######################################################################

# register msgid generated in entity relations tables
_('i18ncard_1'), _('i18ncard_?'), _('i18ncard_+'), _('i18ncard_*')

class CWETypePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('CWEType')
    tabs = [_('cwetype-description'), _('cwetype-box'), _('cwetype-workflow'),
            _('cwetype-views'), _('cwetype-permissions')]
    default_tab = 'cwetype-description'


class CWETypeDescriptionTab(tabs.PrimaryTab):
    __regid__ = 'cwetype-description'
    __select__ = tabs.PrimaryTab.__select__ & is_instance('CWEType')

    def render_entity_attributes(self, entity):
        super(CWETypeDescriptionTab, self).render_entity_attributes(entity)
        _ = self._cw._
        # inheritance
        if entity.specializes:
            self.w(u'<div><strong>%s</strong>' % _('Parent class:'))
            self.wview('csv', entity.related('specializes', 'subject'))
            self.w(u'</div>')
        if entity.reverse_specializes:
            self.w(u'<div><strong>%s</strong>' % _('Sub-classes:'))
            self.wview('csv', entity.related('specializes', 'object'))
            self.w(u'</div>')
        # entity schema image
        self.wview('schemagraph', etype=entity.name)
        # entity schema attributes
        self.w(u'<h2>%s</h2>' % _('CWAttribute_plural'))
        rset = self._cw.execute(
            'Any A,ON,D,C,A,DE,A, IDX,FTI,I18N,R,O,RN,S ORDERBY AA '
            'WHERE A is CWAttribute, A from_entity S, S eid %(x)s, '
            'A ordernum AA, A defaultval D, A description DE, A cardinality C, '
            'A fulltextindexed FTI, A internationalizable I18N, A indexed IDX, '
            'A relation_type R, R name RN, A to_entity O, O name ON',
            {'x': entity.eid})
        self.wview('table', rset, 'null',
                   cellvids={0: 'rdef-name-cell',
                             2: 'etype-attr-defaultval-cell',
                             3: 'etype-attr-cardinality-cell',
                             4: 'rdef-constraints-cell',
                             6: 'rdef-options-cell'},
                   headers=(_(u'name'), _(u'type'),
                            _(u'default value'), _(u'required'),
                            _(u'constraints'), _(u'description'), _('options')))
        # entity schema relations
        self.w(u'<h2>%s</h2>' % _('CWRelation_plural'))
        cellvids = {0: 'rdef-name-cell',
                    2: 'etype-rel-cardinality-cell',
                    3: 'rdef-constraints-cell',
                    4: 'rdef-options-cell'}
        headers= [_(u'name'), _(u'object type'), _(u'cardinality'),
                  _(u'constraints'), _(u'options')]
        rset = self._cw.execute(
            'Any A,TT,"i18ncard_"+SUBSTRING(C,1,1),A,A, K,TTN,R,RN ORDERBY RN '
            'WHERE A is CWRelation, A from_entity S, S eid %(x)s, '
            'A composite K, A cardinality C, '
            'A relation_type R, R name RN, A to_entity TT, TT name TTN',
            {'x': entity.eid})
        if rset:
            self.w(u'<h5>%s %s</h5>' % (entity.name, _('is subject of:')))
            self.wview('table', rset, cellvids=cellvids, headers=headers)
        rset = self._cw.execute(
            'Any A,TT,"i18ncard_"+SUBSTRING(C,1,1),A,A, K,TTN,R,RN ORDERBY RN '
            'WHERE A is CWRelation, A to_entity O, O eid %(x)s, '
            'A composite K, A cardinality C, '
            'A relation_type R, R name RN, A from_entity TT, TT name TTN',
            {'x': entity.eid})
        if rset:
            cellvids[0] = 'rdef-object-name-cell'
            headers[1] = _(u'subject type')
            self.w(u'<h5>%s %s</h5>' % (entity.name, _('is object of:')))
            self.wview('table', rset, cellvids=cellvids, headers=headers)


class CWETypeAttributeCardinalityCell(baseviews.FinalView):
    __regid__ = 'etype-attr-cardinality-cell'

    def cell_call(self, row, col):
        if self.cw_rset.rows[row][col][0] == '1':
            self.w(self._cw._(u'yes'))
        else:
            self.w(self._cw._(u'no'))


class CWETypeAttributeDefaultValCell(baseviews.FinalView):
    __regid__ = 'etype-attr-defaultval-cell'

    def cell_call(self, row, col):
        defaultval = self.cw_rset.rows[row][col]
        if defaultval is not None:
            self.w(text_type(self.cw_rset.rows[row][col].unzpickle()))

class CWETypeRelationCardinalityCell(baseviews.FinalView):
    __regid__ = 'etype-rel-cardinality-cell'

    def cell_call(self, row, col):
        self.w(self._cw._(self.cw_rset.rows[row][col]))


class CWETypeBoxTab(EntityView):
    __regid__ = 'cwetype-box'
    __select__ = is_instance('CWEType')

    def cell_call(self, row, col):
        viewer = schemaviewer.SchemaViewer(self._cw)
        entity = self.cw_rset.get_entity(row, col)
        eschema = self._cw.vreg.schema.eschema(entity.name)
        layout = viewer.visit_entityschema(eschema)
        self.w(uilib.ureport_as_html(layout))
        self.w(u'<br class="clear"/>')


class CWETypePermTab(SecurityViewMixIn, EntityView):
    __regid__ = 'cwetype-permissions'
    __select__ = is_instance('CWEType') & authenticated_user()

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        eschema = self._cw.vreg.schema.eschema(entity.name)
        self.w(u'<h4>%s</h4>' % self._cw._('This entity type permissions:'))
        self.permissions_table(eschema)
        self.w(u'<div style="margin: 0px 1.5em">')
        self.w(u'<h4>%s</h4>' % self._cw._('Attributes permissions:'))
        for attr, etype in  eschema.attribute_definitions():
            if attr not in META_RTYPES:
                rdef = eschema.rdef(attr)
                attrtype = str(rdef.rtype)
                self.w(u'<h4 class="schema">%s (%s)</h4> '
                       % (attrtype, self._cw._(attrtype)))
                self.permissions_table(rdef)
        self.w(u'</div>')


class CWETypeWorkflowTab(EntityView):
    __regid__ = 'cwetype-workflow'
    __select__ = (is_instance('CWEType')
                  & has_related_entities('workflow_of', 'object'))

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        if entity.default_workflow:
            wf = entity.default_workflow[0]
            if len(entity.reverse_workflow_of) > 1:
                self.w(u'<h1>%s (%s)</h1>'
                       % (wf.name, self._cw._('default_workflow')))
            self.display_workflow(wf)
            defaultwfeid = wf.eid
        else:
            self.w(u'<div class="error">%s</div>'
                   % self._cw._('There is no default workflow'))
            defaultwfeid = None
        for altwf in entity.reverse_workflow_of:
            if altwf.eid == defaultwfeid:
                continue
            self.w(u'<h1>%s</h1>' % altwf.name)
            self.display_workflow(altwf)

    def display_workflow(self, wf):
        self.w(wf.view('wfgraph'))
        self.w('<a href="%s">%s</a>' % (
            wf.absolute_url(), self._cw._('more info about this workflow')))


class CWETypeViewsTab(EntityView):
    """possible views for this entity type"""
    __regid__ = 'cwetype-views'
    __select__ = EntityView.__select__ & is_instance('CWEType')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        _ = self._cw._
        self.w('<div>%s</div>' % _('Non exhaustive list of views that may '
                                   'apply to entities of this type'))
        views = [(view.content_type, view.__regid__, _(view.title))
                 for view in self.possible_views(entity.name)]
        self.wview('pyvaltable', pyvalue=sorted(views),
                   headers=(_(u'content type'), _(u'view identifier'),
                            _(u'view title')))

    def possible_views(self, etype):
        rset = self._cw.etype_rset(etype)
        return [v for v in self._cw.vreg['views'].possible_views(self._cw, rset)
                if v.category != 'startupview']


class CWETypeOneLineView(baseviews.OneLineView):
    __select__ = is_instance('CWEType')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        if entity.final:
            self.w(u'<em class="finalentity">')
        super(CWETypeOneLineView, self).cell_call(row, col, **kwargs)
        if entity.final:
            self.w(u'</em>')


# CWRType ######################################################################

class CWRTypePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('CWRType')
    tabs = [_('cwrtype-description'), _('cwrtype-permissions')]
    default_tab = 'cwrtype-description'


class CWRTypeDescriptionTab(tabs.PrimaryTab):
    __regid__ = 'cwrtype-description'
    __select__ = is_instance('CWRType')

    def render_entity_attributes(self, entity):
        super(CWRTypeDescriptionTab, self).render_entity_attributes(entity)
        _ = self._cw._
        if not entity.final:
            self.wview('schemagraph', rtype=entity.name)
        rset = self._cw.execute('Any R,C,R,R, RT WHERE '
                                'R relation_type RT, RT eid %(x)s, '
                                'R cardinality C', {'x': entity.eid})
        self.wview('table', rset, 'null',
                   headers=(_(u'relation'),  _(u'cardinality'), _(u'constraints'),
                            _(u'options')),
                   cellvids={2: 'rdef-constraints-cell',
                             3: 'rdef-options-cell'})


class CWRTypePermTab(SecurityViewMixIn, EntityView):
    __regid__ = 'cwrtype-permissions'
    __select__ = is_instance('CWRType') & authenticated_user()

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rschema = self._cw.vreg.schema.rschema(entity.name)
        self.grouped_permissions_table(rschema)


# CWAttribute / CWRelation #####################################################

class RDEFPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance('CWRelation', 'CWAttribute')
    tabs = [_('rdef-description'), _('rdef-permissions')]
    default_tab = 'rdef-description'


class RDEFDescriptionTab(tabs.PrimaryTab):
    __regid__ = 'rdef-description'
    __select__ = is_instance('CWRelation', 'CWAttribute')

    def render_entity_attributes(self, entity):
        super(RDEFDescriptionTab, self).render_entity_attributes(entity)
        rdef = entity.yams_schema()
        if rdef.constraints:
            self.w(u'<h4>%s</h4>' % self._cw._('constrained_by'))
            self.w(entity.view('rdef-constraints-cell'))


class RDEFPermTab(SecurityViewMixIn, EntityView):
    __regid__ = 'rdef-permissions'
    __select__ = is_instance('CWRelation', 'CWAttribute') & authenticated_user()

    def cell_call(self, row, col):
        self.permissions_table(self.cw_rset.get_entity(row, col).yams_schema())


class RDEFNameView(tableview.CellView):
    """display relation name and its translation only in a cell view, link to
    relation definition's primary view (for use in entity type relations table
    for instance)
    """
    __regid__ = 'rdef-name-cell'
    __select__ = is_instance('CWRelation', 'CWAttribute')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.relation_type[0].name
        # XXX use context entity + pgettext
        self.w(u'<a href="%s">%s</a> (%s)' % (
            entity.absolute_url(), rtype, self._cw._(rtype)))

class RDEFObjectNameView(tableview.CellView):
    """same as RDEFNameView but when the context is the object entity
    """
    __regid__ = 'rdef-object-name-cell'
    __select__ = is_instance('CWRelation', 'CWAttribute')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.relation_type[0].name
        # XXX use context entity + pgettext
        self.w(u'<a href="%s">%s</a> (%s)' % (
            entity.absolute_url(), rtype, self._cw.__(rtype + '_object')))

class RDEFConstraintsCell(EntityView):
    __regid__ = 'rdef-constraints-cell'
    __select__ = is_instance('CWAttribute', 'CWRelation')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rschema = self._cw.vreg.schema.rschema(entity.rtype.name)
        rdef = rschema.rdefs[(entity.stype.name, entity.otype.name)]
        constraints = [xml_escape(text_type(c)) for c in getattr(rdef, 'constraints')]
        self.w(u'<br/>'.join(constraints))

class CWAttributeOptionsCell(EntityView):
    __regid__ = 'rdef-options-cell'
    __select__ = is_instance('CWAttribute')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        options = []
        if entity.indexed:
            options.append(self._cw._('indexed'))
        if entity.fulltextindexed:
            options.append(self._cw._('fulltextindexed'))
        if entity.internationalizable:
            options.append(self._cw._('internationalizable'))
        self.w(u','.join(options))

class CWRelationOptionsCell(EntityView):
    __regid__ = 'rdef-options-cell'
    __select__ = is_instance('CWRelation',)

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.rtype
        options = []
        if rtype.symmetric:
            options.append(self._cw._('symmetric'))
        if rtype.inlined:
            options.append(self._cw._('inlined'))
        if rtype.fulltext_container:
            options.append('%s=%s' % (self._cw._('fulltext_container'),
                                      self._cw._(rtype.fulltext_container)))
        if entity.composite:
            options.append('%s=%s' % (self._cw._('composite'),
                                      self._cw._(entity.composite)))
        self.w(u','.join(options))


# schema images ###############################################################

class RestrictedSchemaVisitorMixIn(object):
    def __init__(self, req, *args, **kwargs):
        self._cw = req
        super(RestrictedSchemaVisitorMixIn, self).__init__(*args, **kwargs)

    def should_display_schema(self, rschema):
        return (super(RestrictedSchemaVisitorMixIn, self).should_display_schema(rschema)
                and rschema.may_have_permission('read', self._cw))

    def should_display_attr(self, eschema, rschema):
        return (super(RestrictedSchemaVisitorMixIn, self).should_display_attr(eschema, rschema)
                and eschema.rdef(rschema).may_have_permission('read', self._cw))


class FullSchemaVisitor(RestrictedSchemaVisitorMixIn, s2d.FullSchemaVisitor):
    pass

class OneHopESchemaVisitor(RestrictedSchemaVisitorMixIn,
                           s2d.OneHopESchemaVisitor):
    pass

class OneHopRSchemaVisitor(RestrictedSchemaVisitorMixIn,
                           s2d.OneHopRSchemaVisitor):
    pass

class CWSchemaDotPropsHandler(s2d.SchemaDotPropsHandler):
    def __init__(self, visitor, cw):
        self.visitor = visitor
        self.cw = cw
        self._cycle = iter(cycle(('#ff7700', '#000000', '#ebbc69', '#888888')))
        self.nextcolor = lambda: next(self._cycle)

        self.colors = {}

    def node_properties(self, eschema):
        """return DOT drawing options for an entity schema include href"""
        label = ['{',eschema.type,'|']
        label.append(r'\l'.join('%s (%s)' % (rel.type, eschema.rdef(rel.type).object)
                                for rel in eschema.ordered_relations()
                                    if rel.final and self.visitor.should_display_attr(eschema, rel)))
        label.append(r'\l}') # trailing \l ensure alignement of the last one
        return {'label' : ''.join(label), 'shape' : "record",
                'fontname' : "Courier", 'style' : "filled",
                'href': self.cw.build_url('cwetype/%s' % eschema.type),
                'fontsize': '10px'
                }

    def edge_properties(self, rschema, subjnode, objnode):
        """return default DOT drawing options for a relation schema"""
        # Inheritance relation (i.e 'specializes').
        if rschema is None:
            kwargs = {'label': 'Parent class',
                      'color' : 'grey',  'style' : 'filled',
                      'arrowhead': 'empty',
                      'fontsize': '10px'}
        # symmetric rels are handled differently, let yams decide what's best
        elif rschema.symmetric:
            kwargs = {'label': rschema.type,
                      'color': '#887788', 'style': 'dashed',
                      'dir': 'both', 'arrowhead': 'normal', 'arrowtail': 'normal',
                      'fontsize': '10px',
                      'href': self.cw.build_url('cwrtype/%s' % rschema.type)}
        else:
            kwargs = {'label': rschema.type,
                      'color' : 'black',  'style' : 'filled', 'fontsize': '10px',
                      'href': self.cw.build_url('cwrtype/%s' % rschema.type)}
            rdef = rschema.rdef(subjnode, objnode)
            composite = rdef.composite
            if rdef.composite == 'subject':
                kwargs['arrowhead'] = 'none'
                kwargs['arrowtail'] = 'diamond'
            elif rdef.composite == 'object':
                kwargs['arrowhead'] = 'diamond'
                kwargs['arrowtail'] = 'none'
            else:
                kwargs['arrowhead'] = 'open'
                kwargs['arrowtail'] = 'none'
            # UML like cardinalities notation, omitting 1..1
            if rdef.cardinality[1] != '1':
                kwargs['taillabel'] = s2d.CARD_MAP[rdef.cardinality[1]]
            if rdef.cardinality[0] != '1':
                kwargs['headlabel'] = s2d.CARD_MAP[rdef.cardinality[0]]
            try:
                kwargs['color'] = self.colors[rschema]
            except KeyError:
                kwargs['color'] = self.nextcolor()
                self.colors[rschema] = kwargs['color']
        kwargs['fontcolor'] = kwargs['color']
        # dot label decoration is just awful (1 line underlining the label
        # + 1 line going to the closest edge spline point)
        kwargs['decorate'] = 'false'
        #kwargs['labelfloat'] = 'true'
        return kwargs


class SchemaGraphView(StartupView):
    __regid__ = 'schemagraph'

    def call(self, etype=None, rtype=None, alt=''):
        if 'MSIE 8' in self._cw.useragent():
            return
        schema = self._cw.vreg.schema
        if etype:
            assert rtype is None
            visitor = OneHopESchemaVisitor(self._cw, schema.eschema(etype),
                                           skiptypes=skip_types(self._cw))
            alt = self._cw._('graphical representation of the %(etype)s '
                             'entity type from %(appid)s data model')
        elif rtype:
            visitor = OneHopRSchemaVisitor(self._cw, schema.rschema(rtype),
                                           skiptypes=skip_types(self._cw))
            alt = self._cw._('graphical representation of the %(rtype)s '
                             'relation type from %(appid)s data model')
        else:
            visitor = FullSchemaVisitor(self._cw, schema,
                                        skiptypes=skip_types(self._cw))
            alt = self._cw._('graphical representation of %(appid)s data model')
        alt %= {'rtype': rtype, 'etype': etype,
                'appid': self._cw.vreg.config.appid}
        prophdlr = CWSchemaDotPropsHandler(visitor, self._cw)
        generator = GraphGenerator(DotBackend('schema', 'BT',
                                              ratio='compress',size=None,
                                              renderer='dot',
                                              additionnal_param={
                                                  'overlap':'false',
                                                  'splines':'true',
                                                  'sep':'0.2',
                                              }))
        # svg image file
        fd, tmpfile = tempfile.mkstemp('.svg')
        try:
            os.close(fd)
            generator.generate(visitor, prophdlr, tmpfile)
            with codecs.open(tmpfile, 'rb', encoding='utf-8') as svgfile:
                self.w(svgfile.read())
        finally:
            os.unlink(tmpfile)

# breadcrumbs ##################################################################

class CWRelationIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('CWRelation')
    def parent_entity(self):
        return self.entity.rtype

class CWAttributeIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('CWAttribute')
    def parent_entity(self):
        return self.entity.stype

class CWConstraintIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('CWConstraint')
    def parent_entity(self):
        if self.entity.reverse_constrained_by:
            return self.entity.reverse_constrained_by[0]

class RQLExpressionIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('RQLExpression')
    def parent_entity(self):
        return self.entity.expression_of


# misc: facets, actions ########################################################

class CWFinalFacet(facet.AttributeFacet):
    __regid__ = 'cwfinal-facet'
    __select__ = facet.AttributeFacet.__select__ & is_instance('CWEType', 'CWRType')
    rtype = 'final'


class ViewSchemaAction(action.Action):
    __regid__ = 'schema'
    __select__ = yes()

    title = _('data model schema')
    order = 30
    category = 'manage'

    def url(self):
        return self._cw.build_url(self.__regid__)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""cubicweb.server.hooks.syncschema unit and functional tests"""

from yams.constraints import BoundaryConstraint

from cubicweb import ValidationError, Binary
from cubicweb.schema import META_RTYPES
from cubicweb.devtools import startpgcluster, stoppgcluster, PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.devtools.repotest import schema_eids_idx


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    stoppgcluster(__file__)
    del SchemaModificationHooksTC.schema_eids


class SchemaModificationHooksTC(CubicWebTC):
    configcls = PostgresApptestConfiguration

    def setUp(self):
        super(SchemaModificationHooksTC, self).setUp()
        self.repo.set_schema(self.repo.deserialize_schema(), resetvreg=False)
        self.__class__.schema_eids = schema_eids_idx(self.repo.schema)

    def index_exists(self, cnx, etype, attr, unique=False):
        dbhelper = self.repo.system_source.dbhelper
        sqlcursor = cnx.cnxset.cu
        return dbhelper.index_exists(sqlcursor,
                                     SQL_PREFIX + etype,
                                     SQL_PREFIX + attr,
                                     unique=unique)

    def _set_perms(self, cnx, eid):
        cnx.execute('SET X read_permission G WHERE X eid %(x)s, G is CWGroup',
                    {'x': eid})
        cnx.execute('SET X add_permission G WHERE X eid %(x)s, G is CWGroup, '
                    'G name "managers"', {'x': eid})
        cnx.execute('SET X delete_permission G WHERE X eid %(x)s, G is CWGroup, '
                    'G name "owners"', {'x': eid})

    def _set_attr_perms(self, cnx, eid):
        cnx.execute('SET X read_permission G WHERE X eid %(x)s, G is CWGroup',
                    {'x': eid})
        cnx.execute('SET X update_permission G WHERE X eid %(x)s, G is CWGroup, G name "managers"',
                    {'x': eid})

    def test_base(self):
        with self.admin_access.repo_cnx() as cnx:
            schema = self.repo.schema
            self.assertFalse(schema.has_entity('Societe2'))
            self.assertFalse(schema.has_entity('concerne2'))
            # schema should be update on insertion (after commit)
            eeid = cnx.execute('INSERT CWEType X: X name "Societe2", '
                               'X description "", X final FALSE')[0][0]
            self._set_perms(cnx, eeid)
            cnx.execute('INSERT CWRType X: X name "concerne2", X description "", '
                        'X final FALSE, X symmetric FALSE')
            self.assertFalse(schema.has_entity('Societe2'))
            self.assertFalse(schema.has_entity('concerne2'))
            # have to commit before adding definition relations
            cnx.commit()
            self.assertTrue(schema.has_entity('Societe2'))
            self.assertTrue(schema.has_relation('concerne2'))
            attreid = cnx.execute('INSERT CWAttribute X: X cardinality "11", '
                                  'X defaultval %(default)s, X indexed TRUE, '
                                  'X relation_type RT, X from_entity E, X to_entity F '
                                  'WHERE RT name "name", E name "Societe2", '
                                  'F name "String"',
                                   {'default': Binary.zpickle('noname')})[0][0]
            self._set_attr_perms(cnx, attreid)
            concerne2_rdef_eid = cnx.execute(
                'INSERT CWRelation X: X cardinality "**", X relation_type RT, '
                'X from_entity E, X to_entity E '
                'WHERE RT name "concerne2", E name "Societe2"')[0][0]
            self._set_perms(cnx, concerne2_rdef_eid)
            self.assertNotIn('name', schema['Societe2'].subject_relations())
            self.assertNotIn('concerne2', schema['Societe2'].subject_relations())
            self.assertFalse(self.index_exists(cnx, 'Societe2', 'name'))
            cnx.commit()
            self.assertIn('name', schema['Societe2'].subject_relations())
            self.assertIn('concerne2', schema['Societe2'].subject_relations())
            self.assertTrue(self.index_exists(cnx, 'Societe2', 'name'))
            # now we should be able to insert and query Societe2
            s2eid = cnx.execute('INSERT Societe2 X: X name "logilab"')[0][0]
            cnx.execute('Societe2 X WHERE X name "logilab"')
            cnx.execute('SET X concerne2 X WHERE X name "logilab"')
            rset = cnx.execute('Any X WHERE X concerne2 Y')
            self.assertEqual(rset.rows, [[s2eid]])
            # check that when a relation definition is deleted, existing relations are deleted
            rdefeid = cnx.execute('INSERT CWRelation X: X cardinality "**", X relation_type RT, '
                                  '   X from_entity E, X to_entity E '
                                  'WHERE RT name "concerne2", E name "CWUser"')[0][0]
            self._set_perms(cnx, rdefeid)
            cnx.commit()
            cnx.execute('DELETE CWRelation X WHERE X eid %(x)s', {'x': concerne2_rdef_eid})
            cnx.commit()
            self.assertIn('concerne2', schema['CWUser'].subject_relations())
            self.assertNotIn('concerne2', schema['Societe2'].subject_relations())
            self.assertFalse(cnx.execute('Any X WHERE X concerne2 Y'))
            # schema should be cleaned on delete (after commit)
            cnx.execute('DELETE CWEType X WHERE X name "Societe2"')
            cnx.execute('DELETE CWRType X WHERE X name "concerne2"')
            self.assertTrue(self.index_exists(cnx, 'Societe2', 'name'))
            self.assertTrue(schema.has_entity('Societe2'))
            self.assertTrue(schema.has_relation('concerne2'))
            cnx.commit()
            self.assertFalse(self.index_exists(cnx, 'Societe2', 'name'))
            self.assertFalse(schema.has_entity('Societe2'))
            self.assertFalse(schema.has_entity('concerne2'))
            self.assertNotIn('concerne2', schema['CWUser'].subject_relations())

    def test_metartype_with_nordefs(self):
        with self.admin_access.repo_cnx() as cnx:
            META_RTYPES.add('custom_meta')
            cnx.execute('INSERT CWRType X: X name "custom_meta", X description "", '
                        'X final FALSE, X symmetric FALSE')
            cnx.commit()
            eeid = cnx.execute('INSERT CWEType X: X name "NEWEtype", '
                               'X description "", X final FALSE')[0][0]
            self._set_perms(cnx, eeid)
            cnx.commit()
            META_RTYPES.remove('custom_meta')

    def test_metartype_with_somerdefs(self):
        with self.admin_access.repo_cnx() as cnx:
            META_RTYPES.add('custom_meta')
            cnx.execute('INSERT CWRType X: X name "custom_meta", X description "", '
                        'X final FALSE, X symmetric FALSE')
            cnx.commit()
            rdefeid = cnx.execute('INSERT CWRelation X: X cardinality "**", X relation_type RT, '
                                  '   X from_entity E, X to_entity E '
                                  'WHERE RT name "custom_meta", E name "CWUser"')[0][0]
            self._set_perms(cnx, rdefeid)
            cnx.commit()
            eeid = cnx.execute('INSERT CWEType X: X name "NEWEtype", '
                               'X description "", X final FALSE')[0][0]
            self._set_perms(cnx, eeid)
            cnx.commit()
            META_RTYPES.remove('custom_meta')

    def test_is_instance_of_insertions(self):
        with self.admin_access.repo_cnx() as cnx:
            seid = cnx.execute('INSERT Transition T: T name "subdiv"')[0][0]
            is_etypes = [etype for etype, in cnx.execute('Any ETN WHERE X eid %s, '
                                                         'X is ET, ET name ETN' % seid)]
            self.assertEqual(is_etypes, ['Transition'])
            instanceof_etypes = [etype
                                 for etype, in cnx.execute('Any ETN WHERE X eid %s, '
                                                           'X is_instance_of ET, ET name ETN'
                                                           % seid)]
            self.assertEqual(sorted(instanceof_etypes), ['BaseTransition', 'Transition'])
            snames = [name for name, in cnx.execute('Any N WHERE S is BaseTransition, S name N')]
            self.assertNotIn('subdiv', snames)
            snames = [name for name, in cnx.execute('Any N WHERE S is_instance_of BaseTransition, '
                                                    'S name N')]
            self.assertIn('subdiv', snames)

    def test_perms_synchronization_1(self):
        with self.admin_access.repo_cnx() as cnx:
            schema = self.repo.schema
            self.assertEqual(schema['CWUser'].get_groups('read'), set(('managers', 'users')))
            self.assertTrue(cnx.execute('Any X, Y WHERE X is CWEType, X name "CWUser", '
                                        'Y is CWGroup, Y name "users"')[0])
            cnx.execute('DELETE X read_permission Y '
                        'WHERE X is CWEType, X name "CWUser", Y name "users"')
            self.assertEqual(schema['CWUser'].get_groups('read'), set(('managers', 'users', )))
            cnx.commit()
            self.assertEqual(schema['CWUser'].get_groups('read'), set(('managers',)))
            cnx.execute('SET X read_permission Y WHERE X is CWEType, '
                        'X name "CWUser", Y name "users"')
            cnx.commit()
            self.assertEqual(schema['CWUser'].get_groups('read'),
                             set(('managers', 'users',)))

    def test_perms_synchronization_2(self):
        with self.admin_access.repo_cnx() as cnx:
            schema = self.repo.schema['in_group'].rdefs[('CWUser', 'CWGroup')]
            self.assertEqual(schema.get_groups('read'),
                             set(('managers', 'users', 'guests')))
            cnx.execute('DELETE X read_permission Y WHERE X relation_type RT, '
                        'RT name "in_group", Y name "guests"')
            self.assertEqual(schema.get_groups('read'),
                             set(('managers', 'users', 'guests')))
            cnx.commit()
            self.assertEqual(schema.get_groups('read'),
                             set(('managers', 'users')))
            cnx.execute('SET X read_permission Y WHERE X relation_type RT, '
                        'RT name "in_group", Y name "guests"')
            self.assertEqual(schema.get_groups('read'),
                             set(('managers', 'users')))
            cnx.commit()
            self.assertEqual(schema.get_groups('read'),
                             set(('managers', 'users', 'guests')))

    def test_nonregr_user_edit_itself(self):
        with self.admin_access.repo_cnx() as cnx:
            ueid = cnx.user.eid
            groupeids = [eid for eid, in cnx.execute('CWGroup G WHERE G name '
                                                     'in ("managers", "users")')]
            cnx.execute('DELETE X in_group Y WHERE X eid %s' % ueid)
            cnx.execute('SET X surname "toto" WHERE X eid %s' % ueid)
            cnx.execute('SET X in_group Y WHERE X eid %s, Y name "managers"' % ueid)
            cnx.commit()
            eeid = cnx.execute('Any X WHERE X is CWEType, X name "CWEType"')[0][0]
            cnx.execute('DELETE X read_permission Y WHERE X eid %s' % eeid)
            cnx.execute('SET X final FALSE WHERE X eid %s' % eeid)
            cnx.execute('SET X read_permission Y WHERE X eid %s, Y eid in (%s, %s)'
                        % (eeid, groupeids[0], groupeids[1]))
            cnx.commit()
            cnx.execute('Any X WHERE X is CWEType, X name "CWEType"')

    # schema modification hooks tests #########################################

    def test_uninline_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            try:
                self.assertTrue(self.schema['state_of'].inlined)
                cnx.execute('SET X inlined FALSE WHERE X name "state_of"')
                self.assertTrue(self.schema['state_of'].inlined)
                cnx.commit()
                self.assertFalse(self.schema['state_of'].inlined)
                self.assertFalse(self.index_exists(cnx, 'State', 'state_of'))
                rset = cnx.execute('Any X, Y WHERE X state_of Y')
                self.assertEqual(len(rset), 2)  # user states
            finally:
                cnx.execute('SET X inlined TRUE WHERE X name "state_of"')
                self.assertFalse(self.schema['state_of'].inlined)
                cnx.commit()
                self.assertTrue(self.schema['state_of'].inlined)
                self.assertTrue(self.index_exists(cnx, 'State', 'state_of'))
                rset = cnx.execute('Any X, Y WHERE X state_of Y')
                self.assertEqual(len(rset), 2)

    def test_indexed_change(self):
        with self.admin_access.repo_cnx() as cnx:
            try:
                cnx.execute('SET X indexed FALSE WHERE X relation_type R, R name "name"')
                self.assertTrue(self.schema['name'].rdef('Workflow', 'String').indexed)
                self.assertTrue(self.index_exists(cnx, 'Workflow', 'name'))
                cnx.commit()
                self.assertFalse(self.schema['name'].rdef('Workflow', 'String').indexed)
                self.assertFalse(self.index_exists(cnx, 'Workflow', 'name'))
            finally:
                cnx.execute('SET X indexed TRUE WHERE X relation_type R, R name "name"')
                self.assertFalse(self.schema['name'].rdef('Workflow', 'String').indexed)
                self.assertFalse(self.index_exists(cnx, 'Workflow', 'name'))
                cnx.commit()
                self.assertTrue(self.schema['name'].rdef('Workflow', 'String').indexed)
                self.assertTrue(self.index_exists(cnx, 'Workflow', 'name'))

    def test_unique_change(self):
        with self.admin_access.repo_cnx() as cnx:
            try:
                eid = cnx.execute('INSERT CWConstraint X: X cstrtype CT, X value "{}", '
                                  '                       DEF constrained_by X '
                                  'WHERE CT name "UniqueConstraint", DEF relation_type RT, '
                                  'DEF from_entity E, RT name "name", '
                                  'E name "Workflow"').rows[0][0]
                self.assertFalse(self.schema['Workflow'].has_unique_values('name'))
                self.assertFalse(self.index_exists(cnx, 'Workflow', 'name', unique=True))
                cnx.commit()
                self.assertTrue(self.schema['Workflow'].has_unique_values('name'))
                self.assertTrue(self.index_exists(cnx, 'Workflow', 'name', unique=True))
            finally:
                cnx.execute('DELETE CWConstraint C WHERE C eid %(eid)s', {'eid': eid})
                cnx.commit()
                self.assertFalse(self.schema['Workflow'].has_unique_values('name'))
                self.assertFalse(self.index_exists(cnx, 'Workflow', 'name', unique=True))

    def test_required_change_1(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('SET DEF cardinality "?1" '
                        'WHERE DEF relation_type RT, DEF from_entity E,'
                        'RT name "title", E name "Bookmark"')
            cnx.commit()
            # should now be able to add bookmark without title
            cnx.execute('INSERT Bookmark X: X path "/view"')
            cnx.commit()

    def test_required_change_2(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('SET DEF cardinality "11" '
                        'WHERE DEF relation_type RT, DEF from_entity E,'
                        'RT name "surname", E name "CWUser"')
            cnx.execute('SET U surname "Doe" WHERE U surname NULL')
            cnx.commit()
            # should not be able anymore to add cwuser without surname
            self.assertRaises(ValidationError, self.create_user, cnx, "toto")
            cnx.rollback()
            cnx.execute('SET DEF cardinality "?1" '
                        'WHERE DEF relation_type RT, DEF from_entity E,'
                        'RT name "surname", E name "CWUser"')
            cnx.commit()

    def test_add_attribute_to_base_class(self):
        with self.admin_access.repo_cnx() as cnx:
            attreid = cnx.execute(
                'INSERT CWAttribute X: X cardinality "11", X defaultval %(default)s, '
                'X indexed TRUE, X relation_type RT, X from_entity E, X to_entity F '
                'WHERE RT name "messageid", E name "BaseTransition", F name "String"',
                {'default': Binary.zpickle('noname')})[0][0]
            assert cnx.execute('SET X read_permission Y WHERE X eid %(x)s, Y name "managers"',
                               {'x': attreid})
            cnx.commit()
            self.schema.rebuild_infered_relations()
            self.assertIn('Transition', self.schema['messageid'].subjects())
            self.assertIn('WorkflowTransition', self.schema['messageid'].subjects())
            cnx.execute('Any X WHERE X is_instance_of BaseTransition, X messageid "hop"')

    def test_change_fulltextindexed(self):
        with self.admin_access.repo_cnx() as cnx:
            target = cnx.create_entity(u'Email', messageid=u'1234',
                                       subject=u'rick.roll@dance.com')
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertIn(target.eid, [item[0] for item in rset])
            assert cnx.execute('SET A fulltextindexed FALSE '
                               'WHERE E is CWEType, E name "Email", A is CWAttribute,'
                               'A from_entity E, A relation_type R, R name "subject"')
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertFalse(rset)
            assert cnx.execute('SET A fulltextindexed TRUE '
                               'WHERE A from_entity E, A relation_type R, '
                               'E name "Email", R name "subject"')
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertIn(target.eid, [item[0] for item in rset])

    def test_change_fulltext_container(self):
        with self.admin_access.repo_cnx() as cnx:
            target = cnx.create_entity(u'EmailAddress', address=u'rick.roll@dance.com')
            target.cw_set(reverse_use_email=cnx.user)
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertIn(cnx.user.eid, [item[0] for item in rset])
            assert cnx.execute('SET R fulltext_container NULL '
                               'WHERE R name "use_email"')
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertIn(target.eid, [item[0] for item in rset])
            assert cnx.execute('SET R fulltext_container "subject" '
                               'WHERE R name "use_email"')
            cnx.commit()
            rset = cnx.execute('Any X WHERE X has_text "rick.roll"')
            self.assertIn(cnx.user.eid, [item[0] for item in rset])

    def test_update_constraint(self):
        with self.admin_access.repo_cnx() as cnx:
            rdef = self.schema['Transition'].rdef('type')
            cstr = rdef.constraint_by_type('StaticVocabularyConstraint')
            cnx.execute('SET X value %(v)s WHERE X eid %(x)s',
                        {'x': cstr.eid, 'v': u"u'normal', u'auto', u'new'"})
            cnx.execute('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, '
                        'EDEF constrained_by X WHERE CT name %(ct)s, EDEF eid %(x)s',
                        {'ct': 'SizeConstraint', 'value': u'max=10', 'x': rdef.eid})
            cnx.commit()
            cstr = rdef.constraint_by_type('StaticVocabularyConstraint')
            self.assertEqual(cstr.values, (u'normal', u'auto', u'new'))
            cnx.execute('INSERT Transition T: T name "hop", T type "new"')

    def test_add_constraint(self):
        with self.admin_access.repo_cnx() as cnx:
            rdef = self.schema['EmailPart'].rdef('ordernum')
            cstr = BoundaryConstraint('>=', 0)
            cnx.execute('INSERT CWConstraint X: X value %(v)s, X cstrtype CT, '
                        'EDEF constrained_by X WHERE CT name %(ct)s, EDEF eid %(x)s',
                        {'ct': cstr.__class__.__name__, 'v': cstr.serialize(), 'x': rdef.eid})
            cnx.commit()
            cstr2 = rdef.constraint_by_type('BoundaryConstraint')
            self.assertEqual(cstr, cstr2)
            cstr3 = BoundaryConstraint('<=', 1000)
            cnx.execute('INSERT CWConstraint X: X value %(v)s, X cstrtype CT, '
                        'EDEF constrained_by X WHERE CT name %(ct)s, EDEF eid %(x)s',
                        {'ct': cstr3.__class__.__name__, 'v': cstr3.serialize(), 'x': rdef.eid})
            cnx.commit()
            # Do not use assertCountEqual as it does "strange" equality
            # comparison on Python 2.
            self.assertEqual(set(rdef.constraints), set([cstr, cstr3]))

    def test_eschema_composite_properties(self):
        with self.admin_access.repo_cnx() as cnx:
            part_eschema = self.schema['EmailPart']
            email_eschema = self.schema['Email']
            parts_rdef = email_eschema.rdef('parts')
            self.assertEqual(part_eschema.composite_rdef_roles, [])
            self.assertEqual(part_eschema.is_composite, False)
            self.assertEqual(email_eschema.composite_rdef_roles,
                             [(parts_rdef, 'subject')])
            self.assertEqual(email_eschema.is_composite, True)
            cnx.execute('DELETE CWRType X WHERE X name "parts"')
            cnx.commit()
            self.assertEqual(part_eschema.composite_rdef_roles, [])
            self.assertEqual(part_eschema.is_composite, False)
            self.assertEqual(email_eschema.composite_rdef_roles, [])
            self.assertEqual(email_eschema.is_composite, False)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Rules based url rewriter component, to get configurable RESTful urls"""

import re

from six import string_types, add_metaclass

from cubicweb.uilib import domid
from cubicweb.appobject import AppObject


def rgx(pattern, flags=0):
    """this is just a convenient shortcut to add the $ sign"""
    return re.compile(pattern+'$', flags)

class metarewriter(type):
    """auto-extend rules dictionary"""
    def __new__(mcs, name, bases, classdict):
        # collect baseclass' rules
        rules = []
        ignore_baseclass_rules = classdict.get('ignore_baseclass_rules', False)
        if not ignore_baseclass_rules:
            for base in bases:
                rules[0:0] = getattr(base, 'rules', [])
        rules[0:0] = classdict.get('rules', [])
        inputs = set()
        for data in rules[:]:
            try:
                input, output, groups = data
            except ValueError:
                input, output = data
            if input in inputs:
                rules.remove( (input, output) )
            else:
                inputs.add(input)
        classdict['rules'] = rules
        return super(metarewriter, mcs).__new__(mcs, name, bases, classdict)


@add_metaclass(metarewriter)
class URLRewriter(AppObject):
    """Base class for URL rewriters.

    Url rewriters should have a `rules` dict that maps an input URI
    to something that should be used for rewriting.

    The actual logic that defines how the rules dict is used is implemented
    in the `rewrite` method.

    A `priority` attribute might be used to indicate which rewriter
    should be tried first. The higher the priority is, the earlier the
    rewriter will be tried.
    """
    __registry__ = 'urlrewriting'
    __abstract__ = True
    priority = 1

    def rewrite(self, req, uri):
        raise NotImplementedError


class SimpleReqRewriter(URLRewriter):
    """The SimpleReqRewriters uses a `rules` dict that maps input URI
    (regexp or plain string) to a dictionary to update the request's
    form.

    If the input uri is a regexp, group substitution is allowed.
    """
    __regid__ = 'simple'

    rules = [
        ('/_', dict(vid='manage')),
        ('/_registry', dict(vid='registry')),
#        (rgx('/_([^/]+?)/?'), dict(vid=r'\1')),
        ('/schema',  dict(vid='schema')),
        ('/index', dict(vid='index')),
        ('/myprefs', dict(vid='propertiesform')),
        ('/siteconfig', dict(vid='systempropertiesform')),
        ('/siteinfo', dict(vid='siteinfo')),
        ('/manage', dict(vid='manage')),
        ('/notfound', dict(vid='404')),
        ('/error', dict(vid='error')),
        ('/sparql', dict(vid='sparql')),
        ('/processinfo', dict(vid='processinfo')),
        (rgx('/cwuser', re.I), dict(vid='cw.users-and-groups-management',
                                    tab=domid('cw.users-management'))),
        (rgx('/cwgroup', re.I), dict(vid='cw.users-and-groups-management',
                                     tab=domid('cw.groups-management'))),
        (rgx('/cwsource', re.I), dict(vid='cw.sources-management')),
        # XXX should be case insensitive as 'create', but I would like to find another way than
        # relying on the etype_selector
        (rgx('/schema/([^/]+?)/?'),  dict(vid='primary', rql=r'Any X WHERE X is CWEType, X name "\1"')),
        (rgx('/add/([^/]+?)/?'), dict(vid='creation', etype=r'\1')),
        (rgx('/doc/images/(.+?)/?'), dict(vid='wdocimages', fid=r'\1')),
        (rgx('/doc/?'), dict(vid='wdoc', fid=r'main')),
        (rgx('/doc/(.+?)/?'), dict(vid='wdoc', fid=r'\1')),
        ]

    def rewrite(self, req, uri):
        """for each `input`, `output `in rules, if `uri` matches `input`,
        req's form is updated with `output`
        """
        for data in self.rules:
            try:
                inputurl, infos, required_groups = data
            except ValueError:
                inputurl, infos = data
                required_groups = None
            if required_groups and not req.user.matching_groups(required_groups):
                continue
            if isinstance(inputurl, string_types):
                if inputurl == uri:
                    req.form.update(infos)
                    break
            elif inputurl.match(uri): # it's a regexp
                # XXX what about i18n? (vtitle for instance)
                for param, value in infos.items():
                    if isinstance(value, string_types):
                        req.form[param] = inputurl.sub(value, uri)
                    else:
                        req.form[param] = value
                break
        else:
            self.debug("no simple rewrite rule found for %s", uri)
            raise KeyError(uri)
        return None, None


def build_rset(rql, rgxgroups=None, setuser=False,
               vid=None, vtitle=None, form={}, **kwargs):

    def do_build_rset(inputurl, uri, req, schema, kwargs=kwargs):
        kwargs = kwargs.copy()
        if rgxgroups:
            match = inputurl.match(uri)
            for arg, group in rgxgroups:
                kwargs[arg] = match.group(group)
        req.form.update(form)
        if setuser:
            kwargs['u'] = req.user.eid
        if vid:
            req.form['vid'] = vid
        if vtitle:
            req.form['vtitle'] = req._(vtitle) % kwargs
        return None, req.execute(rql, kwargs)
    return do_build_rset

def update_form(**kwargs):
    def do_build_rset(inputurl, uri, req, schema):
        match = inputurl.match(uri)
        kwargs.update(match.groupdict())
        req.form.update(kwargs)
        return None, None
    return do_build_rset

def rgx_action(rql=None, args=None, argsgroups=(), setuser=False,
               form=None, formgroups=(), transforms={}, rqlformparams=(), controller=None):
    def do_build_rset(inputurl, uri, req, schema,
                      ):
        if rql:
            kwargs = args and args.copy() or {}
            if argsgroups:
                match = inputurl.match(uri)
                for key in argsgroups:
                    value = match.group(key)
                    try:
                        kwargs[key] = transforms[key](value)
                    except KeyError:
                        kwargs[key] = value
            if setuser:
                kwargs['u'] = req.user.eid
            for param in rqlformparams:
                kwargs.setdefault(param, req.form.get(param))
            rset = req.execute(rql, kwargs)
        else:
            rset = None
        form2 = form and form.copy() or {}
        if formgroups:
            match = inputurl.match(uri)
            for key in formgroups:
                form2[key] = match.group(key)
        if "vtitle" in form2:
            form2['vtitle'] = req.__(form2['vtitle'])
        if form2:
            req.form.update(form2)
        return controller, rset
    return do_build_rset


class SchemaBasedRewriter(URLRewriter):
    """Here, the rules dict maps regexps or plain strings to callbacks
    that will be called with inputurl, uri, req, schema as parameters.
    """
    __regid__ = 'schemabased'
    rules = [
        # rgxp : callback
        (rgx('/search/(.+)'), build_rset(rql=r'Any X ORDERBY FTIRANK(X) DESC WHERE X has_text %(text)s',
                                         rgxgroups=[('text', 1)])),
        ]

    def rewrite(self, req, uri):
        # XXX this could be refacted with SimpleReqRewriter
        for data in self.rules:
            try:
                inputurl, callback, required_groups = data
            except ValueError:
                inputurl, callback = data
                required_groups = None
            if required_groups and not req.user.matching_groups(required_groups):
                continue
            if isinstance(inputurl, string_types):
                if inputurl == uri:
                    return callback(inputurl, uri, req, self._cw.vreg.schema)
            elif inputurl.match(uri): # it's a regexp
                return callback(inputurl, uri, req, self._cw.vreg.schema)
        else:
            self.debug("no schemabased rewrite rule found for %s", uri)
            raise KeyError(uri)



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""CubicWeb web client core. You'll need a apache-modpython or twisted
publisher to get a full CubicWeb web application
"""


from cubicweb import _

from six.moves.urllib.parse import quote as urlquote
from logilab.common.deprecation import deprecated

from cubicweb.web._exceptions import *
from cubicweb.utils import json_dumps
from cubicweb.uilib import eid_param

assert json_dumps is not None, 'no json module installed'

INTERNAL_FIELD_VALUE = '__cubicweb_internal_field__'


class stdmsgs(object):
    """standard ui message (in a class for bw compat)"""
    BUTTON_OK     = (_('button_ok'), 'OK_ICON')
    BUTTON_APPLY  = (_('button_apply'), 'APPLY_ICON')
    BUTTON_CANCEL = (_('button_cancel'), 'CANCEL_ICON')
    BUTTON_DELETE = (_('button_delete'), 'TRASH_ICON')
    YES = (_('yes'), None)
    NO  = (_('no'), None)


from logging import getLogger
LOGGER = getLogger('cubicweb.web')

# XXX deprecated
FACETTES = set()


def jsonize(function):
    def newfunc(*args, **kwargs):
        value = function(*args, **kwargs)
        try:
            return json_dumps(value)
        except TypeError:
            return json_dumps(repr(value))
    return newfunc



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

"""Pyramid resource definitions for CubicWeb."""

from six import text_type

from rql import TypeResolverException

from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPNotFound


class EntityResource(object):

    """A resource class for an entity. It provide method to retrieve an entity
    by eid.
    """

    @classmethod
    def from_eid(cls):
        def factory(request):
            return cls(request, None, None, request.matchdict['eid'])
        return factory

    def __init__(self, request, cls, attrname, value):
        self.request = request
        self.cls = cls
        self.attrname = attrname
        self.value = value

    @reify
    def rset(self):
        req = self.request.cw_request
        if self.cls is None:
            return req.execute('Any X WHERE X eid %(x)s',
                               {'x': int(self.value)})
        st = self.cls.fetch_rqlst(self.request.cw_cnx.user, ordermethod=None)
        st.add_constant_restriction(st.get_variable('X'), self.attrname,
                                    'x', 'Substitute')
        if self.attrname == 'eid':
            try:
                rset = req.execute(st.as_string(), {'x': int(self.value)})
            except (ValueError, TypeResolverException):
                # conflicting eid/type
                raise HTTPNotFound()
        else:
            rset = req.execute(st.as_string(), {'x': text_type(self.value)})
        return rset


class ETypeResource(object):

    """A resource for etype.
    """
    @classmethod
    def from_match(cls, matchname):
        def factory(request):
            return cls(request, request.matchdict[matchname])
        return factory

    def __init__(self, request, etype):
        vreg = request.registry['cubicweb.registry']

        self.request = request
        self.etype = vreg.case_insensitive_etypes[etype.lower()]
        self.cls = vreg['etypes'].etype_class(self.etype)

    def __getitem__(self, value):
        # Try eid first, then rest attribute as for URL path evaluation
        # mecanism in cubicweb.web.views.urlpublishing.
        for attrname in ('eid', self.cls.cw_rest_attr_info()[0]):
            resource = EntityResource(self.request, self.cls, attrname, value)
            try:
                rset = resource.rset
            except HTTPNotFound:
                continue
            if rset.rowcount:
                return resource
        raise KeyError(value)

    @reify
    def rset(self):
        rql = self.cls.fetch_rql(self.request.cw_cnx.user)
        rset = self.request.cw_request.execute(rql)
        return rset



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sources import datafeed


class SynchronizeSourceTC(CubicWebTC):

    def test_synchronize_view(self):
        with self.admin_access.web_request(vid='cw.source-sync') as req:

            class AParser(datafeed.DataFeedParser):
                __regid__ = 'testparser'

                def process(self, url, raise_on_error=False):
                    pass

            with self.temporary_appobjects(AParser):
                source = req.create_entity('CWSource', name=u'ext', type=u'datafeed',
                                           parser=u'cw.entityxml')
                req.cnx.commit()

            self.threads = 0

            def threaded_task(func):
                self.threads += 1

            with tempattr(req.cnx.repo, 'threaded_task', threaded_task):
                path, args = self.expect_redirect_handle_request(
                    req, path=source.rest_path())
                self.assertEqual(self.threads, 1)
                self.assertTrue(path.startswith('cwdataimport/'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.server.sources.storages"""

from six import PY2

from logilab.common.testlib import unittest_main, tag, Tags
from cubicweb.devtools.testlib import CubicWebTC

from glob import glob
import os
import os.path as osp
import sys
import shutil
import tempfile

from cubicweb import Binary, QueryError
from cubicweb.predicates import is_instance
from cubicweb.server.sources import storages
from cubicweb.server.hook import Hook

class DummyBeforeHook(Hook):
    __regid__ = 'dummy-before-hook'
    __select__ = Hook.__select__ & is_instance('File')
    events = ('before_add_entity',)

    def __call__(self):
        self._cw.transaction_data['orig_file_value'] = self.entity.data.getvalue()


class DummyAfterHook(Hook):
    __regid__ = 'dummy-after-hook'
    __select__ = Hook.__select__ & is_instance('File')
    events = ('after_add_entity',)

    def __call__(self):
        # new value of entity.data should be the same as before
        oldvalue = self._cw.transaction_data['orig_file_value']
        assert oldvalue == self.entity.data.getvalue()

class StorageTC(CubicWebTC):
    tempdir = None
    tags = CubicWebTC.tags | Tags('Storage', 'BFSS')

    def setup_database(self):
        self.tempdir = tempfile.mkdtemp()
        bfs_storage = storages.BytesFileSystemStorage(self.tempdir)
        self.bfs_storage = bfs_storage
        storages.set_attribute_storage(self.repo, 'File', 'data', bfs_storage)
        storages.set_attribute_storage(self.repo, 'BFSSTestable', 'opt_attr', bfs_storage)

    def tearDown(self):
        super(StorageTC, self).tearDown()
        storages.unset_attribute_storage(self.repo, 'File', 'data')
        del self.bfs_storage
        shutil.rmtree(self.tempdir)


    def create_file(self, cnx, content=b'the-data'):
        return cnx.create_entity('File', data=Binary(content),
                                 data_format=u'text/plain',
                                 data_name=u'foo.pdf')

    def fspath(self, cnx, entity):
        fspath = cnx.execute('Any fspath(D) WHERE F eid %(f)s, F data D',
                             {'f': entity.eid})[0][0].getvalue()
        return fspath if PY2 else fspath.decode('utf-8')

    def test_bfss_wrong_fspath_usage(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = self.create_file(cnx)
            cnx.execute('Any fspath(D) WHERE F eid %(f)s, F data D', {'f': f1.eid})
            with self.assertRaises(NotImplementedError) as cm:
                cnx.execute('Any fspath(F) WHERE F eid %(f)s', {'f': f1.eid})
            self.assertEqual(str(cm.exception),
                             'This callback is only available for BytesFileSystemStorage '
                             'managed attribute. Is FSPATH() argument BFSS managed?')

    def test_bfss_storage(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            f1 = self.create_file(req)
            filepaths = glob(osp.join(self.tempdir, '%s_data_*' % f1.eid))
            self.assertEqual(len(filepaths), 1, filepaths)
            expected_filepath = filepaths[0]
            # file should be read only
            self.assertFalse(os.access(expected_filepath, os.W_OK))
            self.assertEqual(open(expected_filepath).read(), 'the-data')
            cnx.rollback()
            self.assertFalse(osp.isfile(expected_filepath))
            filepaths = glob(osp.join(self.tempdir, '%s_data_*' % f1.eid))
            self.assertEqual(len(filepaths), 0, filepaths)
            f1 = self.create_file(req)
            cnx.commit()
            filepaths = glob(osp.join(self.tempdir, '%s_data_*' % f1.eid))
            self.assertEqual(len(filepaths), 1, filepaths)
            expected_filepath = filepaths[0]
            self.assertEqual(open(expected_filepath).read(), 'the-data')

            # add f1 back to the entity cache with req as _cw
            f1 = req.entity_from_eid(f1.eid)
            f1.cw_set(data=Binary(b'the new data'))
            cnx.rollback()
            self.assertEqual(open(expected_filepath).read(), 'the-data')
            f1.cw_delete()
            self.assertTrue(osp.isfile(expected_filepath))
            cnx.rollback()
            self.assertTrue(osp.isfile(expected_filepath))
            f1.cw_delete()
            cnx.commit()
            self.assertFalse(osp.isfile(expected_filepath))

    def test_bfss_sqlite_fspath(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = self.create_file(cnx)
            expected_filepath = osp.join(self.tempdir, '%s_data_%s' % (f1.eid, f1.data_name))
            base, ext = osp.splitext(expected_filepath)
            self.assertTrue(self.fspath(cnx, f1).startswith(base))
            self.assertTrue(self.fspath(cnx, f1).endswith(ext))

    def test_bfss_fs_importing_doesnt_touch_path(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.transaction_data['fs_importing'] = True
            filepath = osp.abspath(__file__)
            f1 = cnx.create_entity('File', data=Binary(filepath.encode(sys.getfilesystemencoding())),
                                   data_format=u'text/plain', data_name=u'foo')
            self.assertEqual(self.fspath(cnx, f1), filepath)

    def test_source_storage_transparency(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.temporary_appobjects(DummyBeforeHook, DummyAfterHook):
                self.create_file(cnx)

    def test_source_mapped_attribute_error_cases(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.assertRaises(QueryError) as cm:
                cnx.execute('Any X WHERE X data ~= "hop", X is File')
            self.assertEqual(str(cm.exception), 'can\'t use File.data (X data ILIKE "hop") in restriction')
            with self.assertRaises(QueryError) as cm:
                cnx.execute('Any X, Y WHERE X data D, Y data D, '
                             'NOT X identity Y, X is File, Y is File')
            self.assertEqual(str(cm.exception), "can't use D as a restriction variable")
            # query returning mix of mapped / regular attributes (only file.data
            # mapped, not image.data for instance)
            with self.assertRaises(QueryError) as cm:
                cnx.execute('Any X WITH X BEING ('
                             ' (Any NULL)'
                             '  UNION '
                             ' (Any D WHERE X data D, X is File)'
                             ')')
            self.assertEqual(str(cm.exception), 'query fetch some source mapped attribute, some not')
            with self.assertRaises(QueryError) as cm:
                cnx.execute('(Any D WHERE X data D, X is File)'
                             ' UNION '
                             '(Any D WHERE X title D, X is Bookmark)')
            self.assertEqual(str(cm.exception), 'query fetch some source mapped attribute, some not')

            storages.set_attribute_storage(self.repo, 'State', 'name',
                                           storages.BytesFileSystemStorage(self.tempdir))
            try:
                with self.assertRaises(QueryError) as cm:
                    cnx.execute('Any D WHERE X name D, X is IN (State, Transition)')
                self.assertEqual(str(cm.exception), 'query fetch some source mapped attribute, some not')
            finally:
                storages.unset_attribute_storage(self.repo, 'State', 'name')

    def test_source_mapped_attribute_advanced(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = self.create_file(cnx)
            rset = cnx.execute('Any X,D WITH D,X BEING ('
                                ' (Any D, X WHERE X eid %(x)s, X data D)'
                                '  UNION '
                                ' (Any D, X WHERE X eid %(x)s, X data D)'
                                ')', {'x': f1.eid})
            self.assertEqual(len(rset), 2)
            self.assertEqual(rset[0][0], f1.eid)
            self.assertEqual(rset[1][0], f1.eid)
            self.assertEqual(rset[0][1].getvalue(), b'the-data')
            self.assertEqual(rset[1][1].getvalue(), b'the-data')
            rset = cnx.execute('Any X,LENGTH(D) WHERE X eid %(x)s, X data D',
                                {'x': f1.eid})
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], f1.eid)
            self.assertEqual(rset[0][1], len('the-data'))
            rset = cnx.execute('Any X,LENGTH(D) WITH D,X BEING ('
                                ' (Any D, X WHERE X eid %(x)s, X data D)'
                                '  UNION '
                                ' (Any D, X WHERE X eid %(x)s, X data D)'
                                ')', {'x': f1.eid})
            self.assertEqual(len(rset), 2)
            self.assertEqual(rset[0][0], f1.eid)
            self.assertEqual(rset[1][0], f1.eid)
            self.assertEqual(rset[0][1], len('the-data'))
            self.assertEqual(rset[1][1], len('the-data'))
            with self.assertRaises(QueryError) as cm:
                cnx.execute('Any X,UPPER(D) WHERE X eid %(x)s, X data D',
                             {'x': f1.eid})
            self.assertEqual(str(cm.exception), 'UPPER can not be called on mapped attribute')


    def test_bfss_fs_importing_transparency(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.transaction_data['fs_importing'] = True
            filepath = osp.abspath(__file__)
            f1 = cnx.create_entity('File', data=Binary(filepath.encode(sys.getfilesystemencoding())),
                                   data_format=u'text/plain', data_name=u'foo')
            cw_value = f1.data.getvalue()
            fs_value = open(filepath, 'rb').read()
            if cw_value != fs_value:
                self.fail('cw value %r is different from file content' % cw_value)

    @tag('update')
    def test_bfss_update_with_existing_data(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = cnx.create_entity('File', data=Binary(b'some data'),
                                   data_format=u'text/plain', data_name=u'foo')
            # NOTE: do not use cw_set() which would automatically
            #       update f1's local dict. We want the pure rql version to work
            cnx.execute('SET F data %(d)s WHERE F eid %(f)s',
                         {'d': Binary(b'some other data'), 'f': f1.eid})
            self.assertEqual(f1.data.getvalue(), b'some other data')
            cnx.commit()
            f2 = cnx.execute('Any F WHERE F eid %(f)s, F is File', {'f': f1.eid}).get_entity(0, 0)
            self.assertEqual(f2.data.getvalue(), b'some other data')

    @tag('update', 'extension', 'commit')
    def test_bfss_update_with_different_extension_commited(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = cnx.create_entity('File', data=Binary(b'some data'),
                                   data_format=u'text/plain', data_name=u'foo.txt')
            # NOTE: do not use cw_set() which would automatically
            #       update f1's local dict. We want the pure rql version to work
            cnx.commit()
            old_path = self.fspath(cnx, f1)
            self.assertTrue(osp.isfile(old_path))
            self.assertEqual(osp.splitext(old_path)[1], '.txt')
            cnx.execute('SET F data %(d)s, F data_name %(dn)s, '
                         'F data_format %(df)s WHERE F eid %(f)s',
                         {'d': Binary(b'some other data'), 'f': f1.eid,
                          'dn': u'bar.jpg', 'df': u'image/jpeg'})
            cnx.commit()
            # the new file exists with correct extension
            # the old file is dead
            f2 = cnx.execute('Any F WHERE F eid %(f)s, F is File', {'f': f1.eid}).get_entity(0, 0)
            new_path = self.fspath(cnx, f2)
            self.assertFalse(osp.isfile(old_path))
            self.assertTrue(osp.isfile(new_path))
            self.assertEqual(osp.splitext(new_path)[1], '.jpg')

    @tag('update', 'extension', 'rollback')
    def test_bfss_update_with_different_extension_rolled_back(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = cnx.create_entity('File', data=Binary(b'some data'),
                                   data_format=u'text/plain', data_name=u'foo.txt')
            # NOTE: do not use cw_set() which would automatically
            #       update f1's local dict. We want the pure rql version to work
            cnx.commit()
            old_path = self.fspath(cnx, f1)
            old_data = f1.data.getvalue()
            self.assertTrue(osp.isfile(old_path))
            self.assertEqual(osp.splitext(old_path)[1], '.txt')
            cnx.execute('SET F data %(d)s, F data_name %(dn)s, '
                         'F data_format %(df)s WHERE F eid %(f)s',
                         {'d': Binary(b'some other data'),
                          'f': f1.eid,
                          'dn': u'bar.jpg',
                          'df': u'image/jpeg'})
            cnx.rollback()
            # the new file exists with correct extension
            # the old file is dead
            f2 = cnx.execute('Any F WHERE F eid %(f)s, F is File',
                              {'f': f1.eid}).get_entity(0, 0)
            new_path = self.fspath(cnx, f2)
            new_data = f2.data.getvalue()
            self.assertTrue(osp.isfile(new_path))
            self.assertEqual(osp.splitext(new_path)[1], '.txt')
            self.assertEqual(old_path, new_path)
            self.assertEqual(old_data, new_data)

    @tag('update', 'NULL')
    def test_bfss_update_to_None(self):
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.create_entity('Affaire', opt_attr=Binary(b'toto'))
            cnx.commit()
            f.cw_set(opt_attr=None)
            cnx.commit()

    @tag('fs_importing', 'update')
    def test_bfss_update_with_fs_importing(self):
        with self.admin_access.repo_cnx() as cnx:
            f1 = cnx.create_entity('File', data=Binary(b'some data'),
                                   data_format=u'text/plain',
                                   data_name=u'foo')
            old_fspath = self.fspath(cnx, f1)
            cnx.transaction_data['fs_importing'] = True
            new_fspath = osp.join(self.tempdir, 'newfile.txt')
            open(new_fspath, 'w').write('the new data')
            cnx.execute('SET F data %(d)s WHERE F eid %(f)s',
                         {'d': Binary(new_fspath.encode(sys.getfilesystemencoding())), 'f': f1.eid})
            cnx.commit()
            self.assertEqual(f1.data.getvalue(), b'the new data')
            self.assertEqual(self.fspath(cnx, f1), new_fspath)
            self.assertFalse(osp.isfile(old_fspath))

    @tag('fsimport')
    def test_clean(self):
        with self.admin_access.repo_cnx() as cnx:
            fsimport = storages.fsimport
            td = cnx.transaction_data
            self.assertNotIn('fs_importing', td)
            with fsimport(cnx):
                self.assertIn('fs_importing', td)
                self.assertTrue(td['fs_importing'])
            self.assertNotIn('fs_importing', td)

    @tag('fsimport')
    def test_true(self):
        with self.admin_access.repo_cnx() as cnx:
            fsimport = storages.fsimport
            td = cnx.transaction_data
            td['fs_importing'] = True
            with fsimport(cnx):
                self.assertIn('fs_importing', td)
                self.assertTrue(td['fs_importing'])
            self.assertTrue(td['fs_importing'])

    @tag('fsimport')
    def test_False(self):
        with self.admin_access.repo_cnx() as cnx:
            fsimport = storages.fsimport
            td = cnx.transaction_data
            td['fs_importing'] = False
            with fsimport(cnx):
                self.assertIn('fs_importing', td)
                self.assertTrue(td['fs_importing'])
            self.assertFalse(td['fs_importing'])

if __name__ == '__main__':
    unittest_main()



import webtest

from pyramid.config import Configurator
from cubicweb.devtools.webtest import CubicWebTestTC


class PyramidCWTest(CubicWebTestTC):
    settings = {}

    @classmethod
    def init_config(cls, config):
        super(PyramidCWTest, cls).init_config(config)
        config.global_set_option('anonymous-user', 'anon')

    def setUp(self):
        # Skip CubicWebTestTC setUp
        super(CubicWebTestTC, self).setUp()
        config = Configurator(settings=self.settings)
        config.registry['cubicweb.repository'] = self.repo
        config.include('cubicweb.pyramid')
        self.includeme(config)
        self.pyr_registry = config.registry
        self.webapp = webtest.TestApp(
            config.make_wsgi_app(),
            extra_environ={'wsgi.url_scheme': 'https'})

    def includeme(self, config):
        pass



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

"""Simple statsd_ logger for cubicweb.

This module is meant to be configured by setting a couple of global variables:

- ``bucket`` global variable will be used as statsd bucket in every
statsd_ UDP sent packet.

`- `address`` is a pair (IP, port) specifying the address of the
statsd_ server


There are 3 kinds of statds_ message::

- ``statsd_c(context, n)`` is a simple function to send statsd_
  counter-type of messages like::

    <bucket>.<context>:<n>|c\n

- ``statsd_g(context, value)`` to send statsd_ gauge-type of messages
  like::

    <bucket>.<context>:<n>|g\n

- ``statsd_t(context, ms)`` to send statsd_ time-type of messages
  like::

    <bucket>.<context>:<ms>|ms\n

There is also a decorator (``statsd_timeit``) that may be used to
measure and send to the statsd_ server the time passed in a function
or a method and the number of calls. It will send a message like::
   
    <bucket>.<funcname>:<ms>|ms\n<bucket>.<funcname>:1|c\n


.. _statsd: https://github.com/etsy/statsd

"""



import time
import socket

_bucket = 'cubicweb'
_address = None
_socket = None


def setup(bucket, address):
    """Configure the statsd endpoint

    :param bucket: the name of the statsd bucket that will be used to
                   build messages.

    :param address: the UDP endpoint of the statsd server. Must a
                    couple (ip, port).
    """
    global _bucket, _address, _socket
    packed = None
    for family in (socket.AF_INET6, socket.AF_INET):
        try:
            packed = socket.inet_pton(family, address[0])
            break
        except socket.error:
            continue
    if packed is None:
        return
    _bucket, _address = bucket, address
    _socket = socket.socket(family, socket.SOCK_DGRAM)


def statsd_c(context, n=1):
    if _address is not None:
        _socket.sendto('{0}.{1}:{2}|c\n'.format(_bucket, context, n), _address)


def statsd_g(context, value):
    if _address is not None:
        _socket.sendto('{0}.{1}:{2}|g\n'.format(_bucket, context, value), _address)


def statsd_t(context, value):
    if _address is not None:
        _socket.sendto('{0}.{1}:{2:.4f}|ms\n'.format(_bucket, context, value), _address)


class statsd_timeit(object):
    __slots__ = ('callable',)

    def __init__(self, callableobj):
        self.callable = callableobj

    @property
    def __doc__(self):
        return self.callable.__doc__
    @property
    def __name__(self):
        return self.callable.__name__
    
    def __call__(self, *args, **kw):
        if _address is None:
            return self.callable(*args, **kw)
        t0 = time.time()
        try:
            return self.callable(*args, **kw)
        finally:
            dt = 1000*(time.time()-t0)
            msg = '{0}.{1}:{2:.4f}|ms\n{0}.{1}:1|c\n'.format(_bucket, self.__name__, dt)
            _socket.sendto(msg, _address)
                
    def __get__(self, obj, objtype):
        """Support instance methods."""
        if obj is None: # class method or some already wrapped method
            return self
        import functools
        return functools.partial(self.__call__, obj)



# -*- coding: utf-8 -*-
# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
from contextlib import contextmanager

from logilab.common import tempattr
from logilab.common.testlib import Tags
from cubicweb.devtools.testlib import CubicWebTC

import os
import os.path as osp
import glob

from cubicweb.utils import HTMLHead
from cubicweb.web.views.staticcontrollers import ConcatFilesHandler

class staticfilespublishermixin(object):

    @contextmanager
    def _publish_static_files(self, url, header={}):
        with self.admin_access.web_request(headers=header) as req:
            req._url = url
            self.app_handle_request(req)
            yield req

class StaticControllerCacheTC(staticfilespublishermixin, CubicWebTC):
    tags = CubicWebTC.tags | Tags('static_controller', 'cache', 'http')

    def test_static_file_are_cached(self):
        with self._publish_static_files('data/cubicweb.css') as req:
            self.assertEqual(200, req.status_out)
            self.assertIn('last-modified', req.headers_out)
        next_headers = {
            'if-modified-since': req.get_response_header('last-modified', raw=True),
        }
        with self._publish_static_files('data/cubicweb.css', next_headers) as req:
            self.assertEqual(304, req.status_out)

class StaticDirectoryControllerTC(staticfilespublishermixin, CubicWebTC):

    def test_check_static_dir_access(self):
        """write a file in the static directory and test the access"""
        staticdir = osp.join(self.vreg.config.static_directory)
        if not os.path.exists(staticdir):
            os.makedirs(staticdir)
        filename = osp.join(staticdir, 'test')
        with open(filename, 'a') as f:
            with self._publish_static_files('static/test') as req:
                self.assertEqual(200, req.status_out)

class DataControllerTC(staticfilespublishermixin, CubicWebTC):
    tags = CubicWebTC.tags | Tags('static_controller', 'data', 'http')

    def _check_datafile_ok(self, fname):
        with self._publish_static_files(fname) as req:
            self.assertEqual(200, req.status_out)
            self.assertIn('last-modified', req.headers_out)
            self.assertIn('expires', req.headers_out)
            self.assertEqual(req.get_response_header('cache-control'),
                             {'max-age': 604800})
        next_headers = {
            'if-modified-since': req.get_response_header('last-modified', raw=True),
        }
        with self._publish_static_files(fname, next_headers) as req:
            self.assertEqual(304, req.status_out)

    def _check_datafile_redirect(self, fname, expected):
        with self._publish_static_files(fname) as req:
            self.assertEqual(302, req.status_out)
            self.assertEqual(req.get_response_header('location'),
                             req.base_url() + expected)

    def _check_no_datafile(self, fname):
        with self._publish_static_files(fname) as req:
            self.assertEqual(404, req.status_out)

    def test_static_data_mode(self):
        hash = self.vreg.config.instance_md5_version()
        self.assertEqual(32, len(hash))

        with tempattr(self.vreg.config, 'mode', 'test'):
            self._check_datafile_ok('data/cubicweb.css')
            self._check_no_datafile('data/does/not/exist')
            self._check_no_datafile('data/%s/cubicweb.css' % ('0'*len(hash)))

        with tempattr(self.vreg.config, 'mode', 'notest'):
            self.config._init_base_url()  # reset config.datadir_url
            self._check_datafile_redirect('data/cubicweb.css', 'data/%s/cubicweb.css' % hash)
            self._check_datafile_ok('data/%s/cubicweb.css' % hash)
            self._check_no_datafile('data/%s/does/not/exist' % hash)
            self._check_datafile_redirect('data/%s/does/not/exist' % ('0'*len(hash)),
                                          'data/%s/%s/does/not/exist' % (hash, '0'*len(hash)))


class ConcatFilesTC(CubicWebTC):

    tags = CubicWebTC.tags | Tags('static_controller', 'concat')

    def tearDown(self):
        super(ConcatFilesTC, self).tearDown()
        self._cleanup_concat_cache()

    def _cleanup_concat_cache(self):
        uicachedir = osp.join(self.config.apphome, 'uicache')
        for fname in glob.glob(osp.join(uicachedir, 'cache_concat_*')):
            os.unlink(osp.join(uicachedir, fname))

    @contextmanager
    def _publish_js_files(self, js_files):
        with self.admin_access.web_request() as req:
            head = HTMLHead(req)
            url = head.concat_urls([req.data_url(js_file)
                                    for js_file in js_files])[len(req.base_url()):]
            req._url = url
            res = self.app_handle_request(req)
            yield res, req

    def expected_content(self, js_files):
        content = b''
        for js_file in js_files:
            dirpath, rid = self.config.locate_resource(js_file)
            if dirpath is not None: # ignore resources not found
                with open(osp.join(dirpath, rid), 'rb') as f:
                    content += f.read() + b'\n'
        return content

    def test_cache(self):
        js_files = ('cubicweb.ajax.js', 'jquery.js')
        with self._publish_js_files(js_files) as (result, req):
            self.assertNotEqual(404, req.status_out)
            # check result content
            self.assertEqual(result, self.expected_content(js_files))
            # make sure we kept a cached version on filesystem
            concat_hander = ConcatFilesHandler(self.config)
            filepath = concat_hander.build_filepath(js_files)
            self.assertTrue(osp.isfile(filepath))


    def test_invalid_file_in_debug_mode(self):
        js_files = ('cubicweb.ajax.js', 'dummy.js')
        # in debug mode, an error is raised
        self.config.debugmode = True
        try:
            with self._publish_js_files(js_files) as (result, req):
                #print result
                self.assertEqual(404, req.status_out)
        finally:
            self.config.debugmode = False

    def test_invalid_file_in_production_mode(self):
        js_files = ('cubicweb.ajax.js', 'dummy.js')
        with self._publish_js_files(js_files) as (result, req):
            self.assertNotEqual(404, req.status_out)
            # check result content
            self.assertEqual(result, self.expected_content(js_files))


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""default templates for CubicWeb web client"""


from logilab.mtconverter import xml_escape
from logilab.common.deprecation import class_renamed
from logilab.common.registry import objectify_predicate
from logilab.common.decorators import classproperty

from cubicweb import _
from cubicweb.predicates import match_kwargs, anonymous_user
from cubicweb.view import View, MainTemplate, NOINDEX, NOFOLLOW, StartupView
from cubicweb.utils import UStringIO
from cubicweb.schema import display_name
from cubicweb.web import formfields as ff, formwidgets as fw
from cubicweb.web.views import forms


# main templates ##############################################################

class LogInOutTemplate(MainTemplate):

    def call(self):
        self.set_request_content_type()
        w = self.w
        self.write_doctype()
        self.template_header('text/html', self._cw._('login_action'))
        w(u'<body>\n')
        self.content(w)
        w(u'</body>')

    def template_header(self, content_type, view=None, page_title='', additional_headers=()):
        self._cw.html_headers.define_var('BASE_URL', self._cw.base_url())
        self._cw.html_headers.define_var('DATA_URL', self._cw.datadir_url)
        w = self.whead
        # explictly close the <base> tag to avoid IE 6 bugs while browsing DOM
        w(u'<base href="%s"></base>' % xml_escape(self._cw.base_url()))
        w(u'<meta http-equiv="content-type" content="%s; charset=%s"/>\n'
          % (content_type, self._cw.encoding))
        w(NOINDEX)
        w(NOFOLLOW)
        w(u'\n'.join(additional_headers) + u'\n')
        self.wview('htmlheader', rset=self.cw_rset)
        w(u'<title>%s</title>\n' % xml_escape(page_title))

    def content(self):
        raise NotImplementedError()


class LogInTemplate(LogInOutTemplate):
    __regid__ = 'login'
    __select__ = anonymous_user()
    title = 'log in'

    def content(self, w):
        self.wview('logform', rset=self.cw_rset, id='loginBox', klass='')


class LoggedOutTemplate(StartupView):
    __regid__ = 'loggedout'
    __select__ = anonymous_user()
    title = 'logged out'

    def call(self):
        msg = self._cw._('you have been logged out')
        if self._cw.cnx:
            comp = self._cw.vreg['components'].select('applmessages', self._cw)
            comp.render(w=self.w, msg=msg)
            self.wview('index')
        else:
            self.w(u'<h2>%s</h2>' % msg)


@objectify_predicate
def modal_view(cls, req, rset, *args, **kwargs):
    if req.form.get('__modal', None):
        return 1


@objectify_predicate
def templatable_view(cls, req, rset, *args, **kwargs):
    view = kwargs.pop('view', None)
    if view is None:
        return 1
    if view.binary:
        return 0
    if '__notemplate' in req.form:
        return 0
    return view.templatable


class NonTemplatableViewTemplate(MainTemplate):
    """main template for any non templatable views (xml, binaries, etc.)"""
    __regid__ = 'main-template'
    __select__ = ~templatable_view()

    def call(self, view):
        view.set_request_content_type()
        view.set_stream()
        if (('__notemplate' in self._cw.form)
            and view.templatable
            and view.content_type == self._cw.html_content_type()):
            view.w(u'<div>')
            view.render()
            view.w(u'</div>')
        else:
            view.render()
        # have to replace our stream by view's stream (which may be a binary
        # stream)
        self._stream = view._stream


class ModalMainTemplate(MainTemplate):
    """ a no-decoration main template for standard views
    that typically live in a modal context """
    __regid__ = 'main-template'
    __select__ = templatable_view() & modal_view()

    def call(self, view):
        view.set_request_content_type()
        view.render(w=self.w)


class TheMainTemplate(MainTemplate):
    """default main template :

    - call header / footer templates
    """
    __regid__ = 'main-template'
    __select__ = templatable_view()

    def call(self, view):
        self.set_request_content_type()
        self.template_header(self.content_type, view)
        w = self.w
        w(u'<div id="pageContent">\n')
        vtitle = self._cw.form.get('vtitle')
        if vtitle:
            w(u'<div class="vtitle">%s</div>\n' % xml_escape(vtitle))
        # display entity type restriction component
        etypefilter = self._cw.vreg['components'].select_or_none(
            'etypenavigation', self._cw, rset=self.cw_rset)
        if etypefilter and etypefilter.cw_propval('visible'):
            etypefilter.render(w=w)
        nav_html = UStringIO()
        if view and not view.handle_pagination:
            view.paginate(w=nav_html.write)
        w(nav_html.getvalue())
        w(u'<div id="contentmain">\n')
        view.render(w=w)
        w(u'</div>\n') # close id=contentmain
        w(nav_html.getvalue())
        w(u'</div>\n') # closes id=pageContent
        self.template_footer(view)

    def template_header(self, content_type, view=None, page_title='', additional_headers=()):
        page_title = page_title or view.page_title()
        additional_headers = additional_headers or view.html_headers()
        self.template_html_header(content_type, page_title, additional_headers)
        self.template_body_header(view)

    def template_html_header(self, content_type, page_title, additional_headers=()):
        w = self.whead
        self.write_doctype()
        self._cw.html_headers.define_var('BASE_URL', self._cw.base_url())
        self._cw.html_headers.define_var('DATA_URL', self._cw.datadir_url)
        w(u'<meta http-equiv="content-type" content="%s; charset=%s"/>\n'
          % (content_type, self._cw.encoding))
        w(u'\n'.join(additional_headers) + u'\n')
        self.wview('htmlheader', rset=self.cw_rset)
        if page_title:
            w(u'<title>%s</title>\n' % xml_escape(page_title))

    def template_body_header(self, view):
        w = self.w
        w(u'<body>\n')
        self.wview('header', rset=self.cw_rset, view=view)
        w(u'<div id="page"><table width="100%" border="0" id="mainLayout"><tr>\n')
        self.nav_column(view, 'left')
        w(u'<td id="contentColumn">\n')
        components = self._cw.vreg['components']
        rqlcomp = components.select_or_none('rqlinput', self._cw, rset=self.cw_rset)
        if rqlcomp:
            rqlcomp.render(w=self.w, view=view)
        msgcomp = components.select_or_none('applmessages', self._cw, rset=self.cw_rset)
        if msgcomp:
            msgcomp.render(w=self.w)
        self.content_header(view)

    def template_footer(self, view=None):
        self.content_footer(view)
        self.w(u'</td>\n')
        self.nav_column(view, 'right')
        self.w(u'</tr></table></div>\n')
        self.wview('footer', rset=self.cw_rset, view=view)
        self.w(u'</body>')

    def nav_column(self, view, context):
        boxes = list(self._cw.vreg['ctxcomponents'].poss_visible_objects(
            self._cw, rset=self.cw_rset, view=view, context=context))
        if boxes:
            self.w(u'<td id="navColumn%s"><div class="navboxes">\n' % context.capitalize())
            for box in boxes:
                box.render(w=self.w, view=view)
            self.w(u'</div></td>\n')

    def content_header(self, view=None):
        """by default, display informal messages in content header"""
        self.wview('contentheader', rset=self.cw_rset, view=view)

    def content_footer(self, view=None):
        self.wview('contentfooter', rset=self.cw_rset, view=view)


class ErrorTemplate(TheMainTemplate):
    """fallback template if an internal error occurred during displaying the main
    template. This template may be called for authentication error, which means
    that req.cnx and req.user may not be set.
    """
    __regid__ = 'error-template'

    def call(self):
        """display an unexpected error"""
        self.set_request_content_type()
        self._cw.reset_headers()
        view = self._cw.vreg['views'].select('error', self._cw, rset=self.cw_rset)
        self.template_header(self.content_type, view, self._cw._('an error occurred'),
                             [NOINDEX, NOFOLLOW])
        view.render(w=self.w)
        self.template_footer(view)

    def template_header(self, content_type, view=None, page_title='', additional_headers=()):
        w = self.whead
        self.write_doctype()
        w(u'<meta http-equiv="content-type" content="%s; charset=%s"/>\n'
          % (content_type, self._cw.encoding))
        w(u'\n'.join(additional_headers))
        self.wview('htmlheader', rset=self.cw_rset)
        w(u'<title>%s</title>\n' % xml_escape(page_title))
        self.w(u'<body>\n')

    def template_footer(self, view=None):
        self.w(u'</body>')


class SimpleMainTemplate(TheMainTemplate):

    __regid__ = 'main-no-top'

    def template_header(self, content_type, view=None, page_title='', additional_headers=()):
        page_title = page_title or view.page_title()
        additional_headers = additional_headers or view.html_headers()
        whead = self.whead
        self.write_doctype()
        whead(u'<meta http-equiv="content-type" content="%s; charset=%s"/>\n'
              % (content_type, self._cw.encoding))
        whead(u'\n'.join(additional_headers) + u'\n')
        self.wview('htmlheader', rset=self.cw_rset)
        w = self.w
        whead(u'<title>%s</title>\n' % xml_escape(page_title))
        w(u'<body>\n')
        w(u'<div id="page">')
        w(u'<table width="100%" border="0" id="mainLayout"><tr>\n')
        w(u'<td id="navColumnLeft">\n')
        self.topleft_header()
        boxes = list(self._cw.vreg['ctxcomponents'].poss_visible_objects(
            self._cw, rset=self.cw_rset, view=view, context='left'))
        if boxes:
            w(u'<div class="navboxes">\n')
            for box in boxes:
                box.render(w=w)
            self.w(u'</div>\n')
        w(u'</td>')
        w(u'<td id="contentColumn" rowspan="2">')

    def topleft_header(self):
        logo = self._cw.vreg['components'].select_or_none('logo', self._cw,
                                                          rset=self.cw_rset)
        if logo and logo.cw_propval('visible'):
            w = self.w
            w(u'<table id="header"><tr>\n')
            w(u'<td>')
            logo.render(w=w)
            w(u'</td>\n')
            w(u'</tr></table>\n')


# page parts templates ########################################################

class HTMLHeader(View):
    """default html headers"""
    __regid__ = 'htmlheader'

    def call(self, **kwargs):
        self.favicon()
        self.stylesheets()
        self.javascripts()
        self.alternates()

    def favicon(self):
        favicon = self._cw.uiprops.get('FAVICON', None)
        if favicon:
            self.whead(u'<link rel="shortcut icon" href="%s"/>\n' % favicon)

    def stylesheets(self):
        req = self._cw
        add_css = req.add_css
        for css in req.uiprops['STYLESHEETS']:
            add_css(css, localfile=False)
        for css in req.uiprops['STYLESHEETS_PRINT']:
            add_css(css, u'print', localfile=False)
        for css in req.uiprops['STYLESHEETS_IE']:
            add_css(css, localfile=False, ieonly=True)

    def javascripts(self):
        for jscript in self._cw.uiprops['JAVASCRIPTS']:
            self._cw.add_js(jscript, localfile=False)

    def alternates(self):
        urlgetter = self._cw.vreg['components'].select_or_none('rss_feed_url',
                                                               self._cw, rset=self.cw_rset)
        if urlgetter is not None:
            self.whead(u'<link rel="alternate" type="application/rss+xml" title="RSS feed" href="%s"/>\n'
                       % xml_escape(urlgetter.feed_url()))


class HTMLPageHeader(View):
    """default html page header"""
    __regid__ = 'header'
    main_cell_components = ('appliname', 'breadcrumbs')
    headers = (('headtext', 'header-left'),
               ('header-center', 'header-center'),
               ('header-right', 'header-right')
               )

    def call(self, view, **kwargs):
        self.main_header(view)
        self.w(u'<div id="stateheader">')
        self.state_header()
        self.w(u'</div>')

    def main_header(self, view):
        """build the top menu with authentification info and the rql box"""
        w = self.w
        w(u'<table id="header"><tr>\n')
        for colid, context in self.headers:
            w(u'<td id="%s">' % colid)
            components = self._cw.vreg['ctxcomponents'].poss_visible_objects(
                self._cw, rset=self.cw_rset, view=view, context=context)
            for comp in components:
                comp.render(w=w)
                w(u'&#160;')
            w(u'</td>')
        w(u'</tr></table>\n')

    def state_header(self):
        state = self._cw.search_state
        if state[0] == 'normal':
            return
        _ = self._cw._
        value = self._cw.view('oneline', self._cw.eid_rset(state[1][1]))
        msg = ' '.join((_("searching for"),
                        display_name(self._cw, state[1][3]),
                        _("to associate with"), value,
                        _("by relation"), '"',
                        display_name(self._cw, state[1][2], state[1][0]),
                        '"'))
        return self.w(u'<div class="stateMessage">%s</div>' % msg)


class HTMLPageFooter(View):
    """default html page footer: include footer actions"""
    __regid__ = 'footer'

    def call(self, **kwargs):
        self.w(u'<div id="footer">')
        self.footer_content()
        self.w(u'</div>')

    def footer_content(self):
        actions = self._cw.vreg['actions'].possible_actions(self._cw,
                                                            rset=self.cw_rset,
                                                            view=self.cw_extra_kwargs['view'])
        footeractions = actions.get('footer', ())
        for i, action in enumerate(footeractions):
            self.w(u'<a href="%s">%s</a>' % (action.url(),
                                             self._cw._(action.title)))
            if i < (len(footeractions) - 1):
                self.w(u' | ')


class HTMLContentHeader(View):
    """default html page content header: include selectable content navigation components
    """
    __regid__ = 'contentheader'

    def call(self, view, **kwargs):
        """by default, display informal messages in content header"""
        components = self._cw.vreg['ctxcomponents'].poss_visible_objects(
            self._cw, rset=self.cw_rset, view=view, context='navtop')
        if components:
            self.w(u'<div id="contentheader">')
            for comp in components:
                comp.render(w=self.w, view=view)
            self.w(u'</div><div class="clear"></div>')


class HTMLContentFooter(View):
    """default html page content footer: include selectable content navigation
    components
    """
    __regid__ = 'contentfooter'

    def call(self, view, **kwargs):
        components = self._cw.vreg['ctxcomponents'].poss_visible_objects(
            self._cw, rset=self.cw_rset, view=view, context='navbottom')
        if components:
            self.w(u'<div id="contentfooter">')
            for comp in components:
                comp.render(w=self.w, view=view)
            self.w(u'</div>')


class BaseLogForm(forms.FieldsForm):
    """Abstract Base login form to be used by any login form
    """
    __abstract__ = True

    __regid__ = 'logform'
    domid = 'loginForm'
    needs_css = ('cubicweb.login.css',)

    onclick_base = "javascript: cw.htmlhelpers.popupLoginBox('%s', '%s');"
    onclick_args = (None, None)

    @classproperty
    def form_buttons(cls):
        # we use a property because sub class will need to define their own onclick_args.
        # Therefor we can't juste make the string formating when instanciating this class
        onclick = cls.onclick_base % cls.onclick_args
        form_buttons = [fw.SubmitButton(label=_('log in'),
                                    attrs={'class': 'loginButton'}),
                        fw.ResetButton(label=_('cancel'),
                                       attrs={'class': 'loginButton',
                                              'onclick': onclick}),]
        ## Can't shortcut next access because __dict__ is a "dictproxy" which
        ## does not support items assignement.
        # cls.__dict__['form_buttons'] = form_buttons
        return form_buttons

    def form_action(self):
        if self.action is None:
            # reuse existing redirection if it exists
            target = self._cw.form.get('postlogin_path',
                                       self._cw.relative_path())
            url_args = {}
            if target and target != '/':
                url_args['postlogin_path'] = target
            return self._cw.build_url('login', **url_args)
        return super(BaseLogForm, self).form_action()


class LogForm(BaseLogForm):
    """Simple login form that send username and password
    """
    __regid__ = 'logform'
    domid = 'loginForm'
    needs_css = ('cubicweb.login.css',)
    # XXX have to recall fields name since python is mangling __login/__password
    __login = ff.StringField('__login', widget=fw.TextInput({'class': 'data'}))
    __password = ff.StringField('__password', label=_('password'),
                                widget=fw.PasswordSingleInput({'class': 'data'}))

    onclick_args = ('popupLoginBox', '__login')


class LogFormView(View):
    # XXX an awful lot of hardcoded assumptions there
    #     makes it unobvious to reuse/specialize
    __regid__ = 'logform'
    __select__ = match_kwargs('id', 'klass')

    title = 'log in'

    def call(self, id, klass, title=True, showmessage=True):
        w = self.w
        w(u'<div id="%s" class="%s">' % (id, klass))
        if title:
            stitle = self._cw.property_value('ui.site-title')
            if stitle:
                stitle = xml_escape(stitle)
            else:
                stitle = u'&#160;'
            w(u'<div class="loginTitle">%s</div>' % stitle)
        w(u'<div class="loginContent">\n')
        # don't call self._cw.message twice since it pops the id
        message = self._cw.message
        if showmessage and message:
            w(u'<div class="loginMessage">%s</div>\n' % message)
        config = self._cw.vreg.config
        if config['auth-mode'] != 'http':
            self.login_form(id) # Cookie authentication
        w(u'</div>')
        w(u'</div>\n')

    def login_form(self, id):
        cw = self._cw
        form = cw.vreg['forms'].select('logform', cw)
        if cw.vreg.config['allow-email-login']:
            label = cw._('login or email')
        else:
            label = cw.pgettext('CWUser', 'login')
        form.field_by_name('__login').label = label
        form.render(w=self.w, table_class='', display_progress_div=False)
        cw.html_headers.add_onload('jQuery("#__login:visible").focus()')


LogFormTemplate = class_renamed('LogFormTemplate', LogFormView)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for schema rql (de)serialization"""

from logilab.database import get_db_helper

from yams import register_base_type, unregister_base_type

from cubicweb import Binary
from cubicweb.schema import CubicWebSchemaLoader
from cubicweb import devtools
from cubicweb.devtools.testlib import BaseTestCase as TestCase, CubicWebTC
from cubicweb.server.schemaserial import (updateeschema2rql, updaterschema2rql, rschema2rql,
                                          eschema2rql, rdef2rql, specialize2rql,
                                          _erperms2rql as erperms2rql)


schema = config = None

def setUpModule(*args):
    register_base_type('BabarTestType', ('jungle_speed',))
    helper = get_db_helper('sqlite')
    helper.TYPE_MAPPING['BabarTestType'] = 'TEXT'
    helper.TYPE_CONVERTERS['BabarTestType'] = lambda x: '"%s"' % x

    global schema, config
    loader = CubicWebSchemaLoader()
    config = devtools.TestServerConfiguration('data-schemaserial', __file__)
    config.bootstrap_cubes()
    schema = loader.load(config)


def tearDownModule(*args):
    global schema, config
    schema = config = None

    unregister_base_type('BabarTestType')
    helper = get_db_helper('sqlite')
    helper.TYPE_MAPPING.pop('BabarTestType', None)
    helper.TYPE_CONVERTERS.pop('BabarTestType', None)

cstrtypemap = {'RQLConstraint': 'RQLConstraint_eid',
               'SizeConstraint': 'SizeConstraint_eid',
               'StaticVocabularyConstraint': 'StaticVocabularyConstraint_eid',
               'FormatConstraint': 'FormatConstraint_eid',
               }


class Schema2RQLTC(TestCase):

    def test_eschema2rql1(self):
        self.assertListEqual([
            ('INSERT CWEType X: X description %(description)s,X final %(final)s,X name %(name)s',
             {'description': u'define a final relation: '
              'link a final relation type from a non final entity '
              'to a final entity type. used to build the instance schema',
              'name': u'CWAttribute', 'final': False})],
                             list(eschema2rql(schema.eschema('CWAttribute'))))

    def test_eschema2rql2(self):
        self.assertListEqual([
            ('INSERT CWEType X: X description %(description)s,X final %(final)s,X name %(name)s',
             {'description': u'', 'final': True, 'name': u'String'})],
                             list(eschema2rql(schema.eschema('String'))))

    def test_eschema2rql_specialization(self):
        # x: None since eschema.eid are None
        self.assertListEqual([('SET X specializes ET WHERE X eid %(x)s, ET eid %(et)s',
                               {'et': None, 'x': None}),
                              ('SET X specializes ET WHERE X eid %(x)s, ET eid %(et)s',
                               {'et': None, 'x': None})],
                             sorted(specialize2rql(schema)))

    def test_esche2rql_custom_type(self):
        expected = [('INSERT CWEType X: X description %(description)s,X final %(final)s,'
                     'X name %(name)s',
                     {'description': u'',
                      'name': u'BabarTestType', 'final': True},)]
        got = list(eschema2rql(schema.eschema('BabarTestType')))
        self.assertListEqual(expected, got)

    def test_rschema2rql1(self):
        self.assertListEqual([
            ('INSERT CWRType X: X description %(description)s,X final %(final)s,'
             'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
             'X name %(name)s,X symmetric %(symmetric)s',
             {'description': u'link a relation definition to its relation type',
              'symmetric': False,
              'name': u'relation_type',
              'final': False,
              'fulltext_container': None,
              'inlined': True}),

            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None, 'rt': None, 'oe': None,
              'description': u'',
              'composite': u'object',
              'cardinality': u'1*',
              'ordernum': 1}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None, 'ct': u'RQLConstraint_eid',
              'value': u'{"expression": "O final TRUE", "mainvars": ["O"], "msg": null}'}),

            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None, 'rt': None, 'oe': None,
              'description': u'', 'composite': u'object',
              'ordernum': 1, 'cardinality': u'1*'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None, 'ct': u'RQLConstraint_eid',
              'value': u'{"expression": "O final FALSE", "mainvars": ["O"], "msg": null}'}),
        ],
                             list(rschema2rql(schema.rschema('relation_type'), cstrtypemap)))

    def test_rschema2rql2(self):
        self.assertListEqual([
            ('INSERT CWRType X: X description %(description)s,X final %(final)s,'
             'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
             'X name %(name)s,X symmetric %(symmetric)s',
             {'description': u'',
              'symmetric': False,
              'name': u'add_permission',
              'final': False,
              'fulltext_container': None,
              'inlined': False}),

            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'groups allowed to add entities/relations of this type',
              'composite': None,
              'ordernum': 9999,
              'cardinality': u'**'}),
            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'rql expression allowing to add entities/relations of this type',
              'composite': 'subject',
              'ordernum': 9999,
              'cardinality': u'*?'}),

            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'groups allowed to add entities/relations of this type',
              'composite': None,
              'ordernum': 9999,
              'cardinality': u'**'}),
            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'rql expression allowing to add entities/relations of this type',
              'composite': 'subject',
              'ordernum': 9999,
              'cardinality': u'*?'}),
            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'cardinality': u'**',
              'composite': None,
              'description': u'groups allowed to add entities/relations of this type',
              'oe': None,
              'ordernum': 9999,
              'rt': None,
              'se': None}),
            ('INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,'
             'X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,'
             'X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'cardinality': u'*?',
              'composite': u'subject',
              'description': u'rql expression allowing to add entities/relations of this type',
              'oe': None,
              'ordernum': 9999,
              'rt': None,
              'se': None})],
                             list(rschema2rql(schema.rschema('add_permission'), cstrtypemap)))

    def test_rschema2rql3(self):
        self.assertListEqual([
            ('INSERT CWRType X: X description %(description)s,X final %(final)s,'
             'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
             'X name %(name)s,X symmetric %(symmetric)s',
             {'description': u'',
              'symmetric': False,
              'name': u'cardinality',
              'final': True,
              'fulltext_container': None,
              'inlined': False}),

            ('INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,'
             'X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,'
             'X indexed %(indexed)s,X internationalizable %(internationalizable)s,'
             'X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,'
             'X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'subject/object cardinality',
              'internationalizable': True,
              'fulltextindexed': False,
              'ordernum': 5,
              'defaultval': None,
              'indexed': False,
              'formula': None,
              'cardinality': u'?1'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'ct': u'SizeConstraint_eid',
              'value': u'{"max": 2, "min": null, "msg": null}'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'ct': u'StaticVocabularyConstraint_eid',
              'value': u'{"msg": null, "values": ["?1", "11"]}'}),

            ('INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,'
             'X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,'
             'X indexed %(indexed)s,X internationalizable %(internationalizable)s,'
             'X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,X to_entity OE '
             'WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'subject/object cardinality',
              'internationalizable': True,
              'fulltextindexed': False,
              'ordernum': 5,
              'defaultval': None,
              'indexed': False,
              'formula': None,
              'cardinality': u'?1'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'ct': u'SizeConstraint_eid',
              'value': u'{"max": 2, "min": null, "msg": null}'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'ct': u'StaticVocabularyConstraint_eid',
              "value": (u'{"msg": null, "values": ["?*", "1*", "+*", "**", "?+", "1+", "++", "*+", "?1", '
                        u'"11", "+1", "*1", "??", "1?", "+?", "*?"]}')})],
              list(rschema2rql(schema.rschema('cardinality'), cstrtypemap)))

    def test_rschema2rql_custom_type(self):
        expected = [('INSERT CWRType X: X description %(description)s,X final %(final)s,'
                     'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
                     'X name %(name)s,X symmetric %(symmetric)s',
                     {'description': u'',
                      'final': True,
                      'fulltext_container': None,
                      'inlined': False,
                      'name': u'custom_field_of_jungle',
                      'symmetric': False}),
                     ('INSERT CWAttribute X: X cardinality %(cardinality)s,'
                      'X defaultval %(defaultval)s,X description %(description)s,'
                      'X extra_props %(extra_props)s,X formula %(formula)s,X indexed %(indexed)s,'
                      'X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,'
                      'X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
                      {'cardinality': u'?1',
                       'defaultval': None,
                       'description': u'',
                       'extra_props': b'{"jungle_speed": 42}',
                       'formula': None,
                       'indexed': False,
                       'oe': None,
                       'ordernum': 4,
                       'rt': None,
                       'se': None})]

        got = list(rschema2rql(schema.rschema('custom_field_of_jungle'), cstrtypemap))
        self.assertEqual(2, len(got))
        # this is a custom type attribute with an extra parameter
        self.assertIn('extra_props', got[1][1])
        # this extr
        extra_props = got[1][1]['extra_props']
        self.assertIsInstance(extra_props, Binary)
        got[1][1]['extra_props'] = got[1][1]['extra_props'].getvalue()
        self.assertListEqual(expected, got)

    def test_rdef2rql(self):
        self.assertListEqual([
            ('INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,'
             'X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,'
             'X indexed %(indexed)s,X internationalizable %(internationalizable)s,'
             'X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,'
             'X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s',
             {'se': None,
              'rt': None,
              'oe': None,
              'description': u'',
              'internationalizable': True,
              'fulltextindexed': False,
              'ordernum': 3,
              'defaultval': Binary.zpickle(u'text/plain'),
              'indexed': False,
              'formula': None,
              'cardinality': u'?1'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'value': u'{"msg": null, "values": ["text/rest", "text/markdown", '
              '"text/html", "text/plain"]}',
              'ct': 'FormatConstraint_eid'}),
            ('INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X '
             'WHERE CT eid %(ct)s, EDEF eid %(x)s',
             {'x': None,
              'value': u'{"max": 50, "min": null, "msg": null}',
              'ct': 'SizeConstraint_eid'})],
                             list(rdef2rql(schema['description_format'].rdefs[('CWRType', 'String')],
                                           cstrtypemap)))

    def test_updateeschema2rql1(self):
        self.assertListEqual([('SET X description %(description)s,X final %(final)s,'
                               'X name %(name)s WHERE X eid %(x)s',
                               {'description': u'define a final relation: link a final relation type from'
                                ' a non final entity to a final entity type. used to build the instance schema',
                                'x': 1, 'final': False, 'name': u'CWAttribute'})],
                             list(updateeschema2rql(schema.eschema('CWAttribute'), 1)))

    def test_updateeschema2rql2(self):
        self.assertListEqual([('SET X description %(description)s,X final %(final)s,'
                               'X name %(name)s WHERE X eid %(x)s',
                               {'description': u'', 'x': 1, 'final': True, 'name': u'String'})],
                             list(updateeschema2rql(schema.eschema('String'), 1)))

    def test_updaterschema2rql1(self):
        self.assertListEqual([
            ('SET X description %(description)s,X final %(final)s,'
             'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
             'X name %(name)s,X symmetric %(symmetric)s WHERE X eid %(x)s',
             {'x': 1,
              'symmetric': False,
              'description': u'link a relation definition to its relation type',
              'final': False, 'fulltext_container': None,
              'inlined': True,
              'name': u'relation_type'})],
                             list(updaterschema2rql(schema.rschema('relation_type'), 1)))

    def test_updaterschema2rql2(self):
        expected = [
            ('SET X description %(description)s,X final %(final)s,'
             'X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,'
             'X name %(name)s,X symmetric %(symmetric)s WHERE X eid %(x)s',
             {'x': 1,
              'symmetric': False,
              'description': u'',
              'final': False,
              'fulltext_container': None,
              'inlined': False,
              'name': u'add_permission'})
            ]
        for i, (rql, args) in enumerate(updaterschema2rql(schema.rschema('add_permission'), 1)):
            with self.subTest(i=i):
                self.assertEqual((rql, args), expected[i])


class Perms2RQLTC(TestCase):
    GROUP_MAPPING = {
        'managers': 0,
        'users': 1,
        'guests': 2,
        'owners': 3,
    }

    def test_eperms2rql1(self):
        self.assertListEqual([('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 1}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 2}),
                              ('SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X update_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X delete_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0})],
                             [(rql, kwargs)
                              for rql, kwargs in erperms2rql(schema.eschema('CWEType'), self.GROUP_MAPPING)])

    def test_rperms2rql2(self):
        self.assertListEqual([('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 1}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 2}),
                              ('SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X delete_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0})],
                             [(rql, kwargs)
                              for rql, kwargs in erperms2rql(schema.rschema('read_permission').rdef('CWEType', 'CWGroup'),
                                                             self.GROUP_MAPPING)])

    def test_rperms2rql3(self):
        self.assertListEqual([('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 1}),
                              ('SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 2}),
                              ('SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0}),
                              ('SET X update_permission Y WHERE Y eid %(g)s, X eid %(x)s', {'g': 0})],
                             [(rql, kwargs)
                              for rql, kwargs in erperms2rql(schema.rschema('name').rdef('CWEType', 'String'),
                                                             self.GROUP_MAPPING)])


class ComputedAttributeAndRelationTC(CubicWebTC):
    appid = 'data-cwep002'

    def test(self):
        # force to read schema from the database
        self.repo.set_schema(self.repo.deserialize_schema(), resetvreg=False)
        schema = self.repo.schema
        self.assertEqual([('Company', 'Person')], list(schema['has_employee'].rdefs))
        self.assertEqual(schema['has_employee'].rdef('Company', 'Person').permissions['read'],
                         (u'managers',))
        self.assertEqual('O works_for S',
                         schema['has_employee'].rule)
        self.assertEqual([('Company', 'Int')], list(schema['total_salary'].rdefs))
        self.assertEqual('Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA',
                         schema['total_salary'].rdefs['Company', 'Int'].formula)


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2003-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Views, forms, actions... for the CubicWeb web client"""



import os
import sys
import tempfile

from six import add_metaclass

from rql import nodes
from logilab.mtconverter import xml_escape
from logilab.common.deprecation import class_deprecated


def need_table_view(rset, schema):
    """return True if we think that a table view is more appropriate than a
    list or primary view to display the given result set
    """
    rqlst = rset.syntax_tree()
    if len(rqlst.children) > 1:
        # UNION query, use a table
        return True
    selected = rqlst.children[0].selection
    try:
        mainvar = selected[0]
    except AttributeError:
        # not a variable ref, using table view is probably a good option
        return True
    if not (isinstance(mainvar, nodes.VariableRef) or
            (isinstance(mainvar, nodes.Constant) and mainvar.uid)):
        return True
    for i, etype in enumerate(rset.description[0][1:]):
        # etype may be None on outer join
        if etype is None:
            return True
        # check the selected index node is a VariableRef (else we
        # won't detect aggregate function
        if not isinstance(selected[i+1], nodes.VariableRef):
            return True
        # if this is not a final entity
        if not schema.eschema(etype).final:
            return True
        # if this is a final entity not linked to the main variable
        var = selected[i+1].variable
        for vref in var.references():
            rel = vref.relation()
            if rel is None:
                continue
            if mainvar.is_equivalent(rel.children[0]):
                break
        else:
            return True
    return False

# FIXME: VID_BY_MIMETYPE is unfortunately a bit too naive since
#        some browsers (e.g. FF2) send a bunch of mimetypes in
#        the Accept header, for instance:
#          text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,
#          text/plain;q=0.8,image/png,*/*;q=0.5
VID_BY_MIMETYPE = {
    #'text/xml': 'xml',
    # XXX rss, owl...
}
def vid_from_rset(req, rset, schema, check_table=True):
    """given a result set, return a view id"""
    if rset is None:
        return 'index'
    for mimetype in req.parse_accept_header('Accept'):
        if mimetype in VID_BY_MIMETYPE:
            return VID_BY_MIMETYPE[mimetype]
    nb_rows = len(rset)
    # empty resultset
    if nb_rows == 0:
        return 'noresult'
    # entity result set
    if not schema.eschema(rset.description[0][0]).final:
        if check_table and need_table_view(rset, schema):
            return 'table'
        if nb_rows == 1:
            if req.search_state[0] == 'normal':
                return 'primary'
            return 'outofcontext-search'
        if len(rset.column_types(0)) == 1:
            return 'sameetypelist'
        return 'list'
    return 'table'


def linksearch_select_url(req, rset):
    """when searching an entity to create a relation, return a URL to select
    entities in the given rset
    """
    req.add_js( ('cubicweb.ajax.js', 'cubicweb.edition.js') )
    target, eid, r_type, searchedtype = req.search_state[1]
    if target == 'subject':
        id_fmt = '%s:%s:%%s' % (eid, r_type)
    else:
        id_fmt = '%%s:%s:%s' % (r_type, eid)
    triplets = '-'.join(id_fmt % row[0] for row in rset.rows)
    return "javascript: selectForAssociation('%s', '%s');" % (triplets, eid)


def add_etype_button(req, etype, csscls='addButton right', **urlkwargs):
    vreg = req.vreg
    eschema = vreg.schema.eschema(etype)
    if eschema.has_perm(req, 'add'):
        url = vreg['etypes'].etype_class(etype).cw_create_url(req, **urlkwargs)
        return u'<a href="%s" class="%s">%s</a>' % (
            xml_escape(url), csscls, req.__('New %s' % etype))
    return u''



@add_metaclass(class_deprecated)
class TmpFileViewMixin(object):
    __deprecation_warning__ = '[3.18] %(cls)s is deprecated'
    binary = True
    content_type = 'application/octet-stream'
    cache_max_age = 60*60*2 # stay in http cache for 2 hours by default

    def call(self):
        self.cell_call()

    def cell_call(self, row=0, col=0):
        self.cw_row, self.cw_col = row, col # in case one needs it
        fd, tmpfile = tempfile.mkstemp('.png')
        os.close(fd)
        self._generate(tmpfile)
        self.w(open(tmpfile, 'rb').read())
        os.unlink(tmpfile)
